/* Funcoes de acesso a arquivos de dados sobre 3D.

Derivado de CArquivo.

Formatos de arquivos especificos suportados:

Arquivo de objeto solido (.S):
total_de_vertices     - numero que deve vir antes das secoes
total_de_poligonos    - numero que deve vir antes das secoes, apos total vert
delimitador_de_secao  - define se os valores que seguem sao ou componentes
                        x,y,z de vertices (delimitador eh "V:") ou indices de
                        vertices que compoem poligonos (delimitador: "P:").
                        As secoes podem se alternar: poligonos podem usar
                        vertices definidos em qualquer posicao no arquivo
                        (ateh antes do vertice a ser usado ser definido),
                        pode ocorrer secoes de vertices, poligonos e outras
                        secoes intercaladas.
                        "C:" delimita a secao em que o valor seguinte eh uma
                        cor a ser usada nos poligonos definidos abaixo dessa
                        secao. Cor default: 64. Automaticamente apos o valor
                        da cor ser lido eh alterada a secao para a secao que
                        estava antes de iniciar a secao de cor, a nao ser
                        que um outro delimitador seja encontrado. Essa cor
                        serah a cor base do poligono, cor cujos tons serao
                        usados para definir a sua cor na hora de exibir
                        conforme a tecnica de iluminacao.
                        Secao default no arquivo se solido: "V:"
X1 Y1 Z1              - vertices se for secao de vertices
X2 Y2 Z2
...
VA VB VC VD        - poligonos se for delimitador de secao de poligonos
VE VF VG VH          (V? eh o indice de 0 a total_de_vertices - 1)
                           Obs.: OS VERTICES DEVEM ESTAR DISPOSTOS EM SENTIDO
...                        HORARIO OU ANTI-HORARIO PARA A FUNCAO DE DESENHO
                           DE POLIGONOS FUNCIONAR DIREITO, E DEVEM ESTAR EM
                           SENTIDO ANTI-HORARIO PARA FUNCAO DE REMOCAO DE
                           SUPERFICIES OCULTAS FUNCIONAR!!!
                           (ver CGr::mAtrVrPol2D e CRender::mRetSupVisivel)

Arquivo de particao (.PAR):
total_de_particoes     - deve vir antes de qualquer outra informacao (exceto
                         comentarios).
vertice_inicial_da_particao_1   - indices do 1o. vertice da particao 1
vertice_final_da_particao_1     - indices do ultimo vertice da particao 1
vertice_inicial_da_particao_2   - indices do 1o. vertice da particao 2
vertice_final_da_particao_2     - indices do ultimo vertice da particao 2
...
C: cor poligono_inicial poligono_final - secao de cor, soh possui esses 3
                                         dados. Pode estar intercalado com
                                         particoes. Se nao for definido serah
                                         usada as cores definidas no arquivo
                                         de solido.
... */



#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "defs.h"

#include "arquivo.h"

#include "arq3d.h"

#include "solido.h"

#include "universo.h"



/* construtores - ver no arquivo de header dessa classe
**********************************************************/



/* destrutor
**********************************************************/
CArquivo3D::~CArquivo3D ()
{
}



/* Carrega arquivo de vertices e poligonos que formam solido
**********************************************************/
void CArquivo3D::mRetObjeto ( char* szArq, CSolido& oS )
{
   // abre arquivo de dados
   mAbreArquivo ( szArq, ".s", "rt" );

   // le numero de vertices e poligonos
   UINT16
      uMaxVert = (UINT16) mRetInt (),
      uMaxPol = (UINT16) mRetInt ();

   // aloca memoria
   oS.mAlocaMemoria ( uMaxVert, uMaxPol );

    // le conjunto de vertices e poligonos
   INT16
      nX,
      nY,
      nZ;
   UINT16
      uV1,
      uV2,
      uV3,
      uV4;
   BYTE
      yTipoSec = 'V',
      yCor = 64;
   for ( UINT ucArq = 0; ucArq < ( uMaxVert + uMaxPol ); ucArq ++ ) {

      // le toquen e ve se eh numero ou delimitador de vertices ou poligonos
      char
         szTk [ 200 ];
      mRetStr ( szTk );
      // verifica se eh delimitador de secao de vertices
      if ( strcmp ( "V:", szTk ) == 0 || strcmp ( "v:", szTk ) == 0 ) {
         yTipoSec = 'V';
         // le primeiro x
         nX = mRetInt ();
      }
      else
         // verifica se eh delimitador de secao de poligonos
         if ( strcmp ( "P:", szTk ) == 0 || strcmp ( "p:", szTk ) == 0 ) {
            yTipoSec = 'P';
            // le primeiro vertice
            uV1 = (UINT16) mRetInt ();
         }
         else
            // verifica se eh delimitador de secao de cor
            if ( strcmp ( "C:", szTk ) == 0 || strcmp ( "c:", szTk ) == 0 ) {
               // le cor
               yCor = (BYTE) mRetInt ();
               // procura ateh inicio da proxima secao
               BOOL
                  bFim;
               do {
                  bFim = TRUE;
                  mRetStr ( szTk );
                  // verifica se nao eh nova secao de cor
                  if ( strcmp ( "C:", szTk ) == 0 || strcmp ( "c:", szTk ) == 0 ) {
                     yCor = (BYTE) mRetInt ();
                     bFim = FALSE;
                  }
                  else
                     // secao de poligono
                     if ( strcmp ( "P:", szTk ) == 0 || strcmp ( "p:", szTk ) == 0 ) {
                        yTipoSec = 'P';
                        // le primeiro vertice
                        uV1 = (UINT16) mRetInt ();
                     }
                     else
                        // secao de vertice
                        if ( strcmp ( "V:", szTk ) == 0 || strcmp ( "v:", szTk ) == 0 ) {
                           yTipoSec = 'V';
                           // le primeira componente
                           nX = (UINT16) mRetInt ();
                        }
                        else
                           // numero
                           if ( yTipoSec == 'V' )
                              nX = (INT16) atoi ( szTk );
                           else
                              uV1 = (UINT16) atoi ( szTk );
               } while ( ! bFim );
            }
            else
               // se nao eh delimitador entao eh valor numerico
               // Verifica se valor numerico eh componente x,y,z ou indice de
               // vertice, para isso verifica a qual secao o valor pertence.
               if ( yTipoSec == 'V' )
                  nX = (INT16) atoi ( szTk );
               else
                  uV1 = (UINT16) atoi ( szTk );

      // le demais componentes de vertices ou indices de vertices de poligs.
      // verifica se eh secao de vertices, se sim le componentes
      if ( yTipoSec == 'V' ) {
         nY = mRetInt ();
         nZ = mRetInt ();
         oS.mAtrVertice ( nX, nY, nZ );
      }
      else {
         // le valores como se fossem indices de vertices de poligonos
         uV2 = (UINT16) mRetInt ();
         uV3 = (UINT16) mRetInt ();
         uV4 = (UINT16) mRetInt ();
         oS.mAtrPoligono ( uV1, uV2, uV3, uV4, yCor );
      }
   }

   // fecha arquivo
   mFechaArquivo ();
}



/* Carrega arquivo de particoes que formam solido
**********************************************************/
void CArquivo3D::mRetObjeto ( char* szArq, CUniverso& oU )
{
   // abre arquivo de dados
   mAbreArquivo ( szArq, ".par", "rt" );

   // le numero de partes
   UINT16
      uMaxPa = (UINT16) mRetInt ();

   // aloca memoria
   oU.mAlocaMemoria ( uMaxPa );

   // le particoes
   UINT16
      uIni,
      uFim;
   for ( UINT ucArq = 0; ucArq < uMaxPa; ucArq ++ ) {

      // le toquen e ve se eh numero ou delimitador de secao de cor
      char
         szTk [ 200 ];
      mRetStr ( szTk );
      // se eh delimitador de secao de cor, trata todas as secoes contiguas
      while ( strcmp ( "C:", szTk ) == 0 || strcmp ( "c:", szTk ) == 0 ) {

         // dados da secao de cor
         UINT16
            uCor = (UINT16) mRetInt (),
            uPolIni = (UINT16) mRetInt (),
            uPolFim = (UINT16) mRetInt ();

         // altera tabela de cor dos poligonos, conforme dados da secao
         for ( UINT16 ucP = uPolIni; ucP <= uPolFim; ucP ++ )

            * ( oU.m_payCor + ucP ) = (BYTE) uCor;

         mRetStr ( szTk );
      }

      uIni = (UINT16) atoi ( szTk );
      // le valor como se fosse indice final de vertice da particao
      uFim = (UINT16) mRetInt ();

      oU.mAtrParte ( uIni, uFim );
   }

   // verifica se ainda existe alguma secao de cor para ler
   while ( mProcuraStr ( "C:" ) ) {

      // dados da secao de cor
      UINT16
         uCor = (UINT16) mRetInt (),
         uPolIni = (UINT16) mRetInt (),
         uPolFim = (UINT16) mRetInt ();

      // altera tabela de cor dos poligonos, conforme dados da secao
      for ( UINT16 ucP = uPolIni; ucP <= uPolFim; ucP ++ )
         * ( oU.m_payCor + ucP ) = (BYTE) uCor;
   }

   // fecha arquivo
   mFechaArquivo ();
}
