/* Funcoes de acesso a arquivos.

Formato geral dos arquivos acessaveis:
- arquivos de texto com sequencias de caracteres separadas por espaco(s),
e/ou ENTER(s), e/ou TAB(s). Cada uma dessas sequencias eh denominada token.
Sao possiveis comentarios dentro do corpo do arquivo separados pelos
tokens de inicio e fim de comentario que sao os mesmos do C padrao.
Os tokens de comentario sao balanceados, ou seja, somente termina a area de
comentario quando todos os tokens de abre comentarios tiverem os seus
respectivos fecha comentarios encontrados.
   Exemplos de comentarios: errado -> /*aaa  (emendados), certo -> /* aaa

- arquivos binarios                                   */



#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "defs.h"

#include "arquivo.h"



/* construtor - define sem arquivo corrente
**********************************************************/
CArquivo::CArquivo ()
{
   m_pfArq = NULL;
   m_szNomeArq [ 0 ] = 0;
}



/* construtor - acessa arquivo
**********************************************************/
CArquivo::CArquivo ( char* szNome, char* szExt, char* szModo )
{
   m_pfArq = NULL;
   m_szNomeArq [ 0 ] = 0;
   mAbreArquivo ( szNome, szExt, szModo );
}



/* destrutor - fecha arquivo
**********************************************************/
CArquivo::~CArquivo ()
{
   mFechaArquivo ();
}



/* Abre ou cria arquivo, adiciona extensao ao nome do arquivo e
define niveil inicial de comentarios como 0
szModo - define modo (exemplo: rt, wb). Altera variavel membro
m_pfArq com NULL se houve erro ou com ponteiro para o arquivo.
**********************************************************/
void CArquivo::mAbreArquivo ( char* szNome, char* szExt, char* szModo )
{
   // fecha arquivo atualmente aberto
   mFechaArquivo ();

   // define nivel de comentarios
   m_uNivelComento = 0;

   // adiciona extensao ao nome do arquivo
   strcpy ( m_szNomeArq, szNome );
   strcat ( m_szNomeArq, szExt );

   // tenta abrir arquivo
   if ( ( m_pfArq = fopen ( m_szNomeArq, szModo ) ) == NULL ) {
      printf ( MSG_ERROABRIRARQUIVO, m_szNomeArq );
      exit ( 1 );
   }
}



/* Fecha arquivo se ainda estiver aberto, apaga nome antigo
**********************************************************/
void CArquivo::mFechaArquivo ()
{
   // verifica se arquivo ainda estah aberto
   if ( m_pfArq != NULL ) {
      fclose ( m_pfArq );
      m_pfArq = NULL;
      m_szNomeArq [ 0 ] = 0;  // apaga nome do arquivo antigo
   }
}



/* Verifica se arquivo passado como parametro existe.
return TRUE - sim; FALSE - nao
**********************************************************/
BOOL CArquivo::mRetExisteArquivo ( char* szArq )
{
   FILE
      *pfArq;

   // tenta abrir arquivo, se falhar return FALSE
   if ( ( pfArq = fopen ( szArq, "rb" ) ) == NULL )

      return FALSE;

   // fecha arquivo
   fclose ( pfArq );

   return TRUE;
}



/* Le e altera parametro para token lido de arquivo como string.
Se token for comentario (/*) vai lendo tokens ateh ler o correspondente
"fecha comentario". Os simbolos de comentarios sao balanceados.
Se bAborta TRUE - sai do programa se houve erro, senao nao.
**********************************************************/
BOOL CArquivo::mRetStr ( char* szLinha, BOOL bAborta )
{
   szLinha [ 0 ] = 0;

   BOOL
      bSai;

   // le arquivo ateh achar token que nao esteja entre comentarios
   do {

      bSai = TRUE;

      // verifica se houve fim de comentario se sim fecha comentario
      if ( strcmp ( szLinha, "*/" ) == 0 )
         m_uNivelComento --;

      // tenta ler arquivo, se falhar sai do programa
      if ( fscanf ( m_pfArq, "%s", szLinha ) == 0 || feof ( m_pfArq ) ) {

         // verifica se havia comentario aberto (possivel causa do erro)
         if ( m_uNivelComento > 0 )
            printf ( MSG_ERROCOMENTARIOABERTO, m_uNivelComento );

         if ( bAborta ) {
            printf ( MSG_ERROLEITURASTR );
            exit ( 1 );
         }
         else
            return FALSE;
      }

      // verifica se houve fim de comentario sem inicio correnpondente
      if ( strcmp ( szLinha, "*/" ) == 0 && m_uNivelComento == 0 ) {
         bSai = FALSE;
         printf ( MSG_AVISOFIMCOMENTO );
      }

      // verifica se houve inicio de comentario se sim abre comentario
      if ( strcmp ( szLinha, "/*" ) == 0 )
         m_uNivelComento ++;

   } while ( m_uNivelComento != 0 && bSai );

   return TRUE;
}



/* Le token de arquivo, converte para INT16 e retorna valor.
**********************************************************/
INT16 CArquivo::mRetInt ()
{
   char
      szLinha [ 200 ];

   mRetStr ( szLinha );

   return (INT16) atoi ( szLinha );
}



/* Le uma linha do arquivo. Ignora comentarios.
return - TRUE leu, FALSE - nao leu (fim arquivo talvez)
**********************************************************/
BOOL CArquivo::mRetLinha ( char* szStr, INT nComp )
{
   if ( fgets ( szStr, nComp, m_pfArq ) == NULL )
      return FALSE;

   return TRUE;
}



/* Le um conjunto de bytes (para leitura de arquivo binario)
uNum - numero de bytes a serem lidos
return - TRUE leu, FALSE - nao leu ou leu e achou fim do arquivo
**********************************************************/
BOOL CArquivo::mRetMem ( UINT uNum, void* pvMem )
{
   if ( fread ( pvMem, 1, uNum, m_pfArq ) < uNum )
      return FALSE;

   return TRUE;
}



/* Grava string no arquivo. return TRUE - gravou; FALSE - nao
**********************************************************/
BOOL CArquivo::mAtrStr ( char* szStr )
{
   // tenta gravar arquivo
   if ( fprintf ( m_pfArq, "%s", szStr ) == EOF )
      return FALSE;

   return TRUE;
}



/* Grava um inteiro 16 bits no arquivo. Se bSinal == T
grava com sinal + ou - junto (default) senao grava sem sinal.
return - TRUE gravou, FALSE - nao
**********************************************************/
BOOL CArquivo::mAtrInt ( INT16 nN, BOOL bSinal = TRUE )
{
   char
      szS [ 100 ];

   // verifica se eh para gravar com sinal (se sim grava com sinal)
   if ( bSinal == TRUE )
      sprintf ( szS, "%+06d", nN );
   else
      sprintf ( szS, "%05d", nN );

   return mAtrStr ( szS );
}



/* Grava uma linha em arquivo, poe quedra-de-linha no final.
**********************************************************/
void CArquivo::mAtrLinha ( char* szStr )
{
   fputs ( szStr, m_pfArq );
}



/* Grava um conjunto de bytes (para arquivo binario)
uNum - numero de bytes a serem gravados
return - TRUE gravou, FALSE - nao gravou ou gravou parcialmente
**********************************************************/
BOOL CArquivo::mAtrMem ( UINT uNum, void* pvMem )
{
   if ( fwrite ( pvMem, uNum, 1, m_pfArq ) < 1 )
      return FALSE;

   return TRUE;
}



/* Le arquivo ateh encontrar um string. Termina leitura
no token a ser encontrado. return - TRUE encontrou; FALSE - nao
**********************************************************/
BOOL CArquivo::mProcuraStr ( char* szStr )
{
   char
      szToken [ 200 ];

   // le arquivo ateh achar token procurado
   do
      // verifica se nao chegou ao fim do arquivo
      if ( mRetStr ( szToken, FALSE ) == FALSE )
         return FALSE;
   while ( strcmp ( szToken, szStr ) != 0 );

   return TRUE;
}



/* Posiciona ponteiro de arquivo no inicio do arquivo.
Deslocando conforme nPos.
**********************************************************/
void CArquivo::mPosIni ( INT nPos )
{
   fseek ( m_pfArq, nPos, SEEK_SET );
}



/* Posiciona ponteiro de arquivo a partir da posicao atual.
Deslocando conforme nPos.
**********************************************************/
void CArquivo::mPosAtual ( INT nPos )
{
   fseek ( m_pfArq, nPos, SEEK_CUR );
}



/* Posiciona ponteiro de arquivo no fim do arquivo.
Deslocando conforme nPos.
**********************************************************/
void CArquivo::mPosFim ( INT nPos )
{
   fseek ( m_pfArq, nPos, SEEK_END );
}
