/* Parte de controle das pecas (IA, consistensia,...).

Coordenadas: [linhas][colunas]  de 0 a 7
Pecas: 1 a 12 computador, 13 a 24 usuario,
se peca eh dama: valor = valor OR DAMA.

 posicoes   (0,0)            pecas (valores originais no jogo)
               +*+*+*+*          01  02  03  04
               *+*+*+*+            05  06  07  08    computador
               +*+*+*+*          09  10  11  12
               *+*+*+*+
               +*+*+*+*
               *+*+*+*+          21  22  23  24
               +*+*+*+*            17  18  19  20    usuario
               *+*+*+*+          13  14  15  16
                     (7,7)                                 */



#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "defs.h"

#ifdef TRG_DOS_DJGPP
#include <conio.h>
#include <dpmi.h>

#include "djgpp/teclado.h"
#endif

#include "grvd.h"

#include "controle.h"



#define pu 27
#define du 28
#define pc 29
#define dc 30
#define v 31
#define i 32



// seleciona proxima peca do time do computador
#define SEL_PROX_C \
{ \
   uPSel ++; \
   /* verifica se selecionou peca existe, se nao seleciona primeira */ \
   if ( uPSel >= NPECAS + 1 ) \
      uPSel = 1; \
}



// seleciona proxima peca do time do usuario
#define SEL_PROX_U \
{ \
   uPSel ++; \
   /* verifica se selecionou peca existe, se nao seleciona primeira */ \
   if ( uPSel >= NPECAS + NPECAS + 1 ) \
      uPSel = NPECAS + 1; \
}



// seleciona proximo comando possivel
#define CMD_PROX \
{ \
   yComando ++; \
   /* verifica se selecionou comando pre-programado, se nao seleciona 1o. */ \
   if ( yComando >= 5 ) \
      yComando = 1; \
}



// transforma o codigo de uma tecla no respectivo comando
#define MAPEIA_T_CMD \
   switch ( yComando ) { \
      case T_HOME: \
         yComando = CMD_HOME; \
         break; \
      case T_PG_UP: \
         yComando = CMD_PG_UP;\
         break;\
      case T_END:\
         yComando = CMD_END;\
         break;\
      case T_PG_DN:\
         yComando = CMD_PG_DN;\
         break;\
   }



/* construtor - poe as pecas nas suas posicoes originais,
coerentemente com modelo 3D do tabuleiro.
**********************************************************/
CControle::CControle ()
{
   // limpa o tabuleiro atual
   for ( INT16 ncL = 0; ncL < 8; ncL ++ )
      for ( INT16 ncC = 0; ncC < 8; ncC ++ )
         m_ayTabAtual [ ncL ] [ ncC ] = VAZIO;

   // poe pecas
   // computador (time 1)
   m_ayTabAtual [ 0 ] [ 1 ] = 1;
   m_ayTabAtual [ 0 ] [ 3 ] = 2;
   m_ayTabAtual [ 0 ] [ 5 ] = 3;
   m_ayTabAtual [ 0 ] [ 7 ] = 4;
   m_ayTabAtual [ 1 ] [ 0 ] = 5;
   m_ayTabAtual [ 1 ] [ 2 ] = 6;
   m_ayTabAtual [ 1 ] [ 4 ] = 7;
   m_ayTabAtual [ 1 ] [ 6 ] = 8;
   m_ayTabAtual [ 2 ] [ 1 ] = 9;
   m_ayTabAtual [ 2 ] [ 3 ] = 10;
   m_ayTabAtual [ 2 ] [ 5 ] = 11;
   m_ayTabAtual [ 2 ] [ 7 ] = 12;
   // usuario (time 2)
   m_ayTabAtual [ 7 ] [ 0 ] = 1 + NPECAS;
   m_ayTabAtual [ 7 ] [ 2 ] = 2 + NPECAS;
   m_ayTabAtual [ 7 ] [ 4 ] = 3 + NPECAS;
   m_ayTabAtual [ 7 ] [ 6 ] = 4 + NPECAS;
   m_ayTabAtual [ 6 ] [ 1 ] = 5 + NPECAS;
   m_ayTabAtual [ 6 ] [ 3 ] = 6 + NPECAS;
   m_ayTabAtual [ 6 ] [ 5 ] = 7 + NPECAS;
   m_ayTabAtual [ 6 ] [ 7 ] = 8 + NPECAS;
   m_ayTabAtual [ 5 ] [ 0 ] = 9 + NPECAS;
   m_ayTabAtual [ 5 ] [ 2 ] = 10 + NPECAS;
   m_ayTabAtual [ 5 ] [ 4 ] = 11 + NPECAS;
   m_ayTabAtual [ 5 ] [ 6 ] = 12 + NPECAS;

   // ambas as tabelas comecam com pecas na mesma posicao
   mAtualiza ();
}



/* Exibe tabuleiro em modo texto
**********************************************************/
void CControle::mMostra ()
{
   #ifdef TRG_DOS_DJGPP
   gotoxy ( 1, 1 );
   #endif

   for ( INT16 nL = 0; nL < 8; nL ++ ) {

      for ( INT16 nC = 0; nC < 8; nC ++ )

         // verifica se posicao estah vazia ou nao
         if ( m_ayTabAtual [ nL ] [ nC ] == VAZIO )

            printf ( "  ' " );
         else
            printf ( " %03d", m_ayTabAtual [ nL ] [ nC ] );

      printf ( "\n" );
   }

   #ifdef TRG_DOS_DJGPP
   CTeclado
      oT;
   oT.mRetSoltas ();
   #endif
}



/* Altera tabuleiro atual
**********************************************************/
void CControle::mAtrAtual ( INT16 nValor, INT16 nLin, INT16 nCol )
{
   m_ayTabAtual [ nLin ] [ nCol ] = nValor;
}



/* Copia a tabela de pecas atual para a anterior
**********************************************************/
void CControle::mAtualiza ()
{
   for ( INT16 ncL = 0; ncL < 8; ncL ++ )
      for ( INT16 ncC = 0; ncC < 8; ncC ++ )
         m_ayTabAnter [ ncL ] [ ncC ] = m_ayTabAtual [ ncL ] [ ncC ];
}



/* Define em qual posicao a peca esteve na jogada anterior.
nCol, nLin possuirao a posicao da peca.
return TRUE - achou peca, FALSE - nao achou
**********************************************************/
BOOL CControle::mRetPosAnter ( UINT16 uPeca, INT16& nLin, INT16& nCol )
{
   // varre tabela de posicoes das pecas procurando a peca
   for ( INT16 ncLin = 0; ncLin < 8; ncLin ++ )
      for ( INT16 ncCol = 0; ncCol < 8; ncCol ++ )

         // se a posicao possui a peca define posicao
         if ( m_ayTabAnter [ ncLin ] [ ncCol ] == uPeca ) {

            nCol = ncCol;
            nLin = ncLin;
            return TRUE;
         }

   return FALSE;
}



/* Define em qual posicao a peca estah. nCol, nLin possuirao
a posicao da peca. return TRUE - achou peca, FALSE - nao achou
**********************************************************/
BOOL CControle::mRetPos ( UINT16 uPeca, INT16& nLin, INT16& nCol )
{
   // varre tabela de posicoes das pecas procurando a peca
   for ( INT16 ncLin = 0; ncLin < 8; ncLin ++ )
      for ( INT16 ncCol = 0; ncCol < 8; ncCol ++ )

         // se a posicao possui a peca define posicao
         if ( m_ayTabAtual [ ncLin ] [ ncCol ] == uPeca ) {

            nCol = ncCol;
            nLin = ncLin;
            return TRUE;
         }

   return FALSE;
}



/* Verifica que time estah ocupando uma posicao.
return TIMEU - peca eh do time do usuario, TIMEC - computador,
VAZIO - posicao desocupara.
**********************************************************/
BYTE CControle::mRetPosTime ( INT16 nLin, INT16 nCol )
{
   UINT16
      uPeca = mRetPeca ( nLin, nCol );

   // converte peca de dama para valor original
   uPeca = mRetValorOrigPeca ( uPeca );

   // verifica se peca pertence ao time do computador
   if ( uPeca >= 1 && uPeca <= 12 )
      return TIMEC;

   // verifica se peca pertence ao time do usuario
   if ( uPeca >= 1 + NPECAS && uPeca <= 12 + NPECAS )
      return TIMEU;

   return VAZIO;
}



/* Retorna a peca existente numa determinada posicao ou VAZIO
 se a posicao estah desocupada. Valor muda se a peca for dama.
**********************************************************/
INT16 CControle::mRetPeca ( INT16 nLin, INT16 nCol )
{
   // bugfix - GPF
   if ( nLin < 0 || nCol < 0 )
      return VAZIO;
      
   return m_ayTabAtual [ nLin ] [ nCol ];
}



/* 1 - Remove a peca da tabela de posicoes do tabuleiro.
Eh especificada a peca.
**********************************************************/
void CControle::mRemove ( UINT16 uPeca )
{
   INT16
      nCol,
      nLin;

   // enquanto encontrar ocorrencias da peca
   BOOL
      bFim = FALSE;
   do {
      // se encontrou a peca a apaga
      if ( mRetPos ( uPeca, nLin, nCol ) )
         m_ayTabAtual [ nLin ] [ nCol ] = VAZIO;
      else
         // se nao encontrou a peca para de procurar
         bFim = TRUE;

   } while ( ! bFim );
}



/* 2 - Remove a peca da tabela de posicoes do tabuleiro.
Eh especificada a posicao da peca.
**********************************************************/
void CControle::mRemove ( INT16 nLin, INT16 nCol )
{
   m_ayTabAtual [ nLin ] [ nCol ] = VAZIO;
}



/* Coloca uma peca numa determinada posicao do tabuleiro.
Apaga a peca da sua posicao anterior.
 Verifica se coordenadas nao estao fora do tabuleiro.
 return TRUE - feito, FALSE - nao feito (fora do tabuleiro)
**********************************************************/
BOOL CControle::mColoca ( UINT16 uPeca, INT16 nLin, INT16 nCol )
{
   // verifica se linha esta dentro do tabuleiro
   if ( nLin >= 0 && nLin < 8 )
      // verifica se coluna esta dentro do tabuleiro
      if ( nCol >= 0 && nCol < 8 ) {

         // altera peca de posicao
         mRemove ( uPeca );
         m_ayTabAtual [ nLin ] [ nCol ] = uPeca;
         return TRUE;
      }

   return FALSE;
}



/* Desloca relativamente a peca de acordo com a sua posicao atual.
 Verifica se a posicao jah esta ocupada por alguma peca e se movimento
 sai do tabuleiro.
 return TRUE - movimento feito,
        FALSE - nao feito (sai do tabuleiro ou posicao jah ocupada)
**********************************************************/
BOOL CControle::mMove ( UINT16 uPeca, INT16 nDLin, INT16 nDCol )
{
   // define posicao da peca
   INT16
      nCol,
      nLin;
   mRetPos ( uPeca, nLin, nCol );

   // verifica se posicao destino jah estah ocupada
   if ( mRetPosTime ( nLin + nDLin, nCol + nDCol ) != VAZIO )
      return FALSE;

   // tenta colocar peca na posicao
   return mColoca ( uPeca, nLin + nDLin, nCol + nDCol );
}



/* Transforma peca em dama. uPeca informa qual peca transformar.
**********************************************************/
void CControle::mDamifica ( UINT16 uPeca )
{
   INT16
      nCol,
      nLin;

   mRetPos ( uPeca, nLin, nCol );
   mColoca ( uPeca | DAMA, nLin, nCol );
}



/* Retorna valor atual de uma peca. uPeca indica valor
original da peca no jogo, retorna valor atual da peca.
Se a peca for dama retorna seu valor como dama, se for
peca normal retorna seu valor como peca normal, se estiver
morta retorna VAZIO.
Funcao boa para fazer referencia a pecas em seu estado normal
ou como damas (elimina preocupacao com seu estado atual).
**********************************************************/
UINT16 CControle::mRetValorAtualPeca ( UINT16 uPeca )
{
   INT16
      nCol,
      nLin;
   UINT16
      uValor = VAZIO;

   // verifica se peca estah com seu valor original
   if ( mRetPos ( uPeca, nLin, nCol ) == TRUE )
      uValor = uPeca;

   // verifica se peca estah com seu valor de dama
   if ( mRetPos ( uPeca | DAMA, nLin, nCol ) == TRUE )
      uValor = uPeca | DAMA;

   return uValor;
}



/* Retorna valor original de uma peca. uPeca indica valor
atual da peca no jogo, retorna valor no inicio do jogo da peca.
Se a peca for dama retorna seu valor como peca comum, se for
peca normal retorna seu valor como peca normal, se estiver
morta retorna VAZIO.
Funcao boa para fazer referencia a pecas em seu estado normal
desconvertendo de damas.
**********************************************************/
UINT16 CControle::mRetValorOrigPeca ( UINT16 uPeca )
{
   INT16
      nCol,
      nLin;
   UINT16
      uValor = VAZIO;

   // verifica se peca estah com seu valor de dama
   if ( mRetPos ( uPeca | DAMA, nLin, nCol ) == TRUE )
      uValor = uPeca - DAMA;
   else
      // verifica se peca estah com seu valor original
      if ( mRetPos ( uPeca, nLin, nCol ) == TRUE )
         uValor = uPeca;

   return uValor;
}



/* Verifica se peca esta viva, mesmo sendo dama.
**********************************************************/
BOOL CControle::mRetViva ( UINT16 uPeca )
{
   INT16
      nCol,
      nLin;

   if ( mRetValorAtualPeca ( uPeca ) == VAZIO )
      return FALSE;

   return TRUE;
}



/* Verifica se peca eh dama
**********************************************************/
BOOL CControle::mRetEhDama ( UINT16 uPeca )
{
   INT16
      nCol,
      nLin;

   return mRetPos ( uPeca | DAMA, nLin, nCol );
}



/* Verifica se uma peca deve atacar. uPeca deve ser
o valor original da peca no jogo.
return TRUE - sim, FALSE - nao;
ayDirecao - direcoes para onde pode atacar (relativo ao usuario)
Se nDir == 1:
   posicao no vetor 0 - frente esquerda, 1 - frente direita,
                    2 - tras   esquerda, 3 - tras   direita.
Se nDir == -1:
   posicao no vetor 0 - tras   direita, 1 - tras   esquerda,
                    2 - frente direita, 3 - frente esquerda.
Se peca eh dama eh testado se pode atacar para tras.
nDir informa direcao de avanco/ataque dos peoes.
**********************************************************/
BOOL CControle::mRetDeveAtacar ( UINT16 uPeca, INT16 nDir, BYTE* ayDirecoes )
{
   // define posicao atual da peca
   INT16
      nCol,
      nLin;
   mRetPos ( mRetValorAtualPeca ( uPeca ), nLin, nCol );

   // deve atacar
   BOOL
      bAtacar = FALSE;

   // verifica se direcao de ataque eh do usuario, se sim testa para usuario
   if ( nDir == 1 ) {

      // verifica se peca pode atacar para frente esquerda
      if ( nLin > 1 && nCol > 1 &&               // testa limites do tabuleiro
           mRetPosTime ( nLin - 1 * nDir, nCol - 1 * nDir ) == TIMEC &&  // testa time
           mRetPosTime ( nLin - 2 * nDir, nCol - 2 * nDir ) == VAZIO ) { // testa vazio

         ayDirecoes [ 0 ] = TRUE;
         bAtacar = TRUE;
      }

      // verifica se peca pode atacar para frente direita
      if ( nLin > 1 && nCol < 6 &&               // testa limites do tabuleiro
           mRetPosTime ( nLin - 1 * nDir, nCol + 1 * nDir ) == TIMEC &&  // testa time
           mRetPosTime ( nLin - 2 * nDir, nCol + 2 * nDir ) == VAZIO ) { // testa vazio

         ayDirecoes [ 1 ] = TRUE;
         bAtacar = TRUE;
      }

      // testa se a peca eh dama, se for nao deve testar se deve comer para tras
      if ( mRetEhDama ( uPeca ) ) {

         // verifica se peca pode atacar para tras esquerda
         if ( nLin < 6 && nCol > 1 &&               // testa limites do tabuleiro
              mRetPosTime ( nLin + 1 * nDir, nCol - 1 * nDir ) == TIMEC &&  // testa time
              mRetPosTime ( nLin + 2 * nDir, nCol - 2 * nDir ) == VAZIO ) { // testa vazio

            ayDirecoes [ 2 ] = TRUE;
            bAtacar = TRUE;
         }

         // verifica se peca pode atacar para tras direita
         if ( nLin < 6 && nCol < 6 &&               // testa limites do tabuleiro
              mRetPosTime ( nLin + 1 * nDir, nCol + 1 * nDir ) == TIMEC &&  // testa time
              mRetPosTime ( nLin + 2 * nDir, nCol + 2 * nDir ) == VAZIO ) { // testa vazio

            ayDirecoes [ 3 ] = TRUE;
            bAtacar = TRUE;
         }
      }
   }
   else {
   // testa para computador (nDir == -1), direcoes relativo ao computador

      // verifica se peca pode atacar para frente esquerda
      if ( nLin < 6 && nCol < 6 &&              // testa limites do tabuleiro
           mRetPosTime ( nLin - 1 * nDir, nCol - 1 * nDir ) == TIMEU &&  // time
           mRetPosTime ( nLin - 2 * nDir, nCol - 2 * nDir ) == VAZIO ) { // vazio

         ayDirecoes [ 0 ] = TRUE;
         bAtacar = TRUE;
      }

      // verifica se peca pode atacar para frente direita
      if ( nLin < 6 && nCol > 1 &&              // testa limites do tabuleiro
           mRetPosTime ( nLin - 1 * nDir, nCol + 1 * nDir ) == TIMEU &&  // time
           mRetPosTime ( nLin - 2 * nDir, nCol + 2 * nDir ) == VAZIO ) { // vazio

         ayDirecoes [ 1 ] = TRUE;
         bAtacar = TRUE;
      }

      // testa se a peca eh dama, se for nao deve testar se deve comer para tras
      if ( mRetEhDama ( uPeca ) ) {

         // verifica se peca pode atacar para tras esquerda
         if ( nLin > 1 && nCol < 6 &&           // testa limites do tabuleiro
              mRetPosTime ( nLin + 1 * nDir, nCol - 1 * nDir ) == TIMEU &&  // time
              mRetPosTime ( nLin + 2 * nDir, nCol - 2 * nDir ) == VAZIO ) { // vazio

            ayDirecoes [ 2 ] = TRUE;
            bAtacar = TRUE;
         }

         // verifica se peca pode atacar para tras direita
         if ( nLin > 1 && nCol > 1 &&           // testa limites do tabuleiro
              mRetPosTime ( nLin + 1 * nDir, nCol + 1 * nDir ) == TIMEU &&  // time
              mRetPosTime ( nLin + 2 * nDir, nCol + 2 * nDir ) == VAZIO ) { // vazio

            ayDirecoes [ 3 ] = TRUE;
            bAtacar = TRUE;
         }
      }
   }

   return bAtacar;
}



/* Se possivel faz uma jogada com a peca selecionada.
return 0 (ou FALSE) - jogada invalida, comando para direcao nao permitida;
       1 (ou TRUE)  - fez jogada e terminou;
       2            - fez jogada, mas estah incompleta (ataque em sequencia);
       3            - jogada invalida: outra peca estah em posicao para comer
Se outra peca estah em posicao de ataque seleciona essa peca e return 3.
uPeca eh o valor original da peca, mas internamente eh
transformada para o valor atual da peca (dependente de ser dama)
e deve ser devolvida como valor original da peca (sem considerar dama).
nDir informa a direcao de avanco das pecas, se
1 - avancar significa mover pecas do lado do usuario para
o lado do adversario, se -1 - avancar eh mover pecas do
computador para o usuario. Direcao de yComando depende de nDir.
**********************************************************/
BOOL CControle::mMonoMove ( BYTE yComando, INT16 nDir, UINT16& uPeca )
{
   // transforma em dama, se peca eh dama
   if ( mRetEhDama ( uPeca ) )
      uPeca = uPeca | DAMA;

   // define numero de casas que deverao se puladas de uma vez
   // originalmente 1 casa - movimento simples, 2 - ataque
   INT16
      nPulo = 1 * nDir;

   // se comando valido
   BOOL
      yResMov = 1,       // se jogada terminou
      bAtacar = FALSE;   // se atacando (tomando uma peca adversaria)

   // verifica se peca selecionada deve atacar
   BYTE
      ayDirecao [ 4 ] = { FALSE, FALSE, FALSE, FALSE };
   if ( mRetDeveAtacar ( uPeca, nDir, ayDirecao ) ) {

      // verifica se a direcao comandada eh uma das direcoes possiveis
      if ( ( yComando == CMD_HOME  && ayDirecao [ 0 ] ) ||   // frente esquerda
           ( yComando == CMD_PG_UP && ayDirecao [ 1 ] ) ||    // frente direita
           ( yComando == CMD_END   && ayDirecao [ 2 ] ) ||    // tras esquerda
           ( yComando == CMD_PG_DN && ayDirecao [ 3 ] ) ) {  // tras direita

         nPulo = 2 * nDir;
         bAtacar = TRUE;
      }
      else {

         // tranforma peca para valor original
         if ( mRetEhDama ( uPeca ) )
            uPeca -= DAMA;
         // comando nao eh para uma das direcoes permitidas
         return 0;
      }
   }
   else {
      // define de qual time eh a peca
      INT16
         nTime = TIMEU;
      if ( nDir == -1 )
         nTime = TIMEC;

      // procura alguma peca que deve atacar
      for ( UINT16 ucP = 1 + nTime; ucP <= 12 + nTime; ucP ++ )

         // verifica se peca estah viva e deve atacar
         if ( mRetViva ( ucP ) && mRetDeveAtacar ( ucP, nDir, ayDirecao ) ) {

            // define essa como sendo peca selecionada
            uPeca = mRetValorAtualPeca ( ucP );

            // tranforma peca para valor original
            if ( mRetEhDama ( uPeca ) )
               uPeca -= DAMA;

            return 3;
         }
   }

   // move ou ataca removendo peca atacada
   // verifica se a peca selecionada eh dama
   if ( mRetEhDama ( uPeca ) ) {

      // verifica qual foi o comando do usuario
      if ( yComando == CMD_HOME ) {

         yResMov = mMove ( uPeca, - nPulo, - nPulo );

         // remove peca atacada
         if ( yResMov == 1 && bAtacar ) {
            INT16
               nCol,
               nLin;
            mRetPos ( uPeca, nLin, nCol );
            mRemove ( nLin + nDir, nCol + nDir );
         }
      }

      if ( yComando == CMD_PG_UP ) {

         yResMov = mMove ( uPeca, - nPulo,   nPulo );

         // remove peca atacada
         if ( yResMov == 1 && bAtacar ) {
            INT16
               nCol,
               nLin;
            mRetPos ( uPeca, nLin, nCol );
            mRemove ( nLin + nDir, nCol - nDir );
         }
      }

      if ( yComando == CMD_END ) {

         yResMov = mMove ( uPeca,   nPulo, - nPulo );

         // remove peca atacada
         if ( yResMov == 1 && bAtacar ) {
            INT16
               nCol,
               nLin;
            mRetPos ( uPeca, nLin, nCol );
            mRemove ( nLin - nDir, nCol + nDir );
         }
      }

      if ( yComando == CMD_PG_DN ) {

         yResMov = mMove ( uPeca,   nPulo,   nPulo );

         // remove peca atacada
         if ( yResMov == 1 && bAtacar ) {
            INT16
               nCol,
               nLin;
            mRetPos ( uPeca, nLin, nCol );
            mRemove ( nLin - nDir, nCol - nDir );
         }
      }

      // verifica se deve continuar jogada: fazer ataque em sequencia
      if ( bAtacar && mRetDeveAtacar ( uPeca, nDir, ayDirecao ) )
         yResMov = 2;

      // tranforma peca para valor original
      uPeca -= DAMA;
   }
   else {
      // se peca nao eh dama
      // verifica qual foi o comando do usuario
      if ( yComando == CMD_HOME ) {

         yResMov = mMove ( uPeca, - nPulo, - nPulo );

         // remove peca atacada
         if ( yResMov == 1 && bAtacar ) {
            INT16
               nCol,
               nLin;
            mRetPos ( uPeca, nLin, nCol );
            mRemove ( nLin + nDir, nCol + nDir );
         }
      }

      if ( yComando == CMD_PG_UP ) {

         yResMov = mMove ( uPeca, - nPulo,   nPulo );

         // remove peca atacada
         if ( yResMov == 1 && bAtacar ) {
            INT16
               nCol,
               nLin;
            mRetPos ( uPeca, nLin, nCol );
            mRemove ( nLin + nDir, nCol - nDir );
         }
      }

      // comando nao eh para uma direcao permitida
      if ( yComando == CMD_END || yComando == CMD_PG_DN )
         return 0;

      // verifica se deve continuar jogada: fazer ataque em sequencia
      if ( bAtacar && mRetDeveAtacar ( uPeca, nDir, ayDirecao ) && // atacar
           ( ayDirecao [ 0 ] || ayDirecao [ 1 ] ) )   // direcoes para frente
         yResMov = 2;
   }

   return yResMov;
}



/* Faz uma sequencia de movimentos consistente que completa
 uma jogada (come em sequencia e se tem alguma peca que deve
 comer entao joga com essa peca). Transforma peca em dama se for o caso.
Sempre move alguma peca, se nao for possivel a selecionada tenta uma outra.
yComando - direcao que a principio deve ser usada no ataque
nTime - time da peca (TIMEC ou TIMEU)
uPSel - valor original da peca que, a principio, deve ser usada no movimento
return 0 (FALSE) - termina partida;
       1 (TRUE)  - continua partida;
Termina jogo se: nao existem pecas para mover, todas pecas estao bloqueadas.
**********************************************************/
BOOL CControle::mAutoMoveObrig ( BYTE yComando, INT16 nTime, UINT16& uPSel )
{
   // define direcao de movimentacao para frente conforme time
   INT16
      nDir;
   // se time eh do computador
   if ( nTime == TIMEC )
      nDir = -1;
   else
      // se time eh do usuario
      nDir = 1;

   // verifica se peca estah viva, se nao estah seleciona uma seguinte
   INT16
      ncLimiteProc = 0;
   BOOL
      bFimProc = FALSE;
   while ( ! mRetViva ( uPSel ) && ! bFimProc ) {
      // seleciona peca
      if ( nTime == TIMEC )
         SEL_PROX_C
      else
         SEL_PROX_U

      // verifica se terminou jogo (nao existem pecas)
      // verifica se jah procurou todas pecas possiveis
      if ( ncLimiteProc > NPECAS + 1 )
         return FALSE;
      else
         // conta pecas jah procuradas
         ncLimiteProc ++;
   }

   INT16
      nResMov1,     // resultado do primeiro movimento tentado
      nResMov2 = 5; // resultado dos demais movimentos que ocorreram se
                    // selecionou outra peca ou estah em ataque em sequencia

   // enquanto nao selecionar uma peca que poca mover e fizer movimento tenta
   BOOL
      bFimMov1;
   INT16
      ncLimiteMov1 = 0;
   do {

      // assume que peca serah movida
      bFimMov1 = TRUE;

      // joga com a peca selecionada ou seleciona uma peca que deve comer
      nResMov1 = mMonoMove ( yComando, nDir, uPSel );

      // verifica se peca selecionada pode ir para outra direcao
      // se peca nao se moveu e nao existe nenhuma outra em posicao de ataque
      if ( nResMov1 == 0 ) {

         // tenta ir para alguma direcao se ainda nao conseguiu mover
         if ( nResMov2 == 0 || nResMov2 == 5 )
            nResMov2 = mMonoMove ( 0, nDir, uPSel );
         if ( nResMov2 == 0 || nResMov2 == 5 )
            nResMov2 = mMonoMove ( 1, nDir, uPSel );
         if ( nResMov2 == 0 || nResMov2 == 5 )
            nResMov2 = mMonoMove ( 2, nDir, uPSel );
         if ( nResMov2 == 0 || nResMov2 == 5 )
            nResMov2 = mMonoMove ( 3, nDir, uPSel );

         // se nao fez movimento para qualquer direcao tenta outra peca
         if ( nResMov2 == 0 ) {
            if ( nTime == TIMEC )
               SEL_PROX_C
            else
               SEL_PROX_U

            // faz mais um ciclo de tentativas com proxima peca
            bFimMov1 = FALSE;
         }
      }

      // se tentou todas as pecas e nao conseguiu desiste: fim do jogo
      if ( ncLimiteMov1 == 12 )
         return FALSE;
      else
         // conta mais uma tentativa
         ncLimiteMov1 ++;

   } while ( ! bFimMov1 );

   // verifica se selecionou outra peca que deve comer, se sim joga essa peca
   if ( nResMov1 == 3 ) {

      // tenta atacar para cada uma das direcoes, quando conseguir aborta
      for ( INT16 ncCmd = 0; ncCmd < 5; ncCmd ++ )
         if ( ( nResMov2 = mMonoMove ( ncCmd, nDir, uPSel ) ) != 0 )
            break;
   }

   // verifica se algum dos 2 movimentos acima formou uma sequencia
   if ( nResMov1 == 2 || nResMov2 == 2 ) {

      BOOL
         bFimMove = FALSE;
      // enquanto nao terminar sequencia executa
      while ( ! bFimMove ) {

         // tenta atacar para alguma direcao
         nResMov2 = mMonoMove ( 0, nDir, uPSel );
         if ( nResMov2 == 0 )
            nResMov2 = mMonoMove ( 1, nDir, uPSel );
         if ( nResMov2 == 0 )
            nResMov2 = mMonoMove ( 2, nDir, uPSel );
         if ( nResMov2 == 0 )
            nResMov2 = mMonoMove ( 3, nDir, uPSel );

         // verifica se sequencia continua ou se terminou
         if ( nResMov2 != 2 )
            // terminou sequencia
            bFimMove = TRUE;
      }
   }

   // verifica se peca estah em posicao de se tornar dama, se sim transforma
   // verifica se peca jah nao eh dama
   if ( ! mRetEhDama ( uPSel ) )
      // se peca eh do time do usuario, verifica linha 0
      if ( nTime == TIMEU ) {
         // verifica se peca estah na linha extrema
         INT16
            nCol,
            nLin;
         mRetPos ( uPSel, nLin, nCol );
         if ( nLin == 0 )
            // transforma em dama
            mDamifica ( uPSel );
      }
      else
      // se peca eh do time do computador, verifica linha 7
      if ( nTime == TIMEC ) {
         // verifica se peca estah na linha extrema
         INT16
            nCol,
            nLin;
         mRetPos ( uPSel, nLin, nCol );
         if ( nLin == 7 )
            // transforma em dama
            mDamifica ( uPSel );
      }

   return TRUE;
}



/* Copia o tabuleiro atual para um tabuleiro temporario local.
**********************************************************/
void CControle::mMSalvaAtual ( BYTE* ayTab )
{
   memcpy ( (void*) ayTab, (const void*) m_ayTabAtual, TAMTAB );
}



/* Copia tabuleiro temporario local para tabuleiro atual.
**********************************************************/
void CControle::mMRestauraAtual ( BYTE* ayTab )
{
   memcpy ( (void*) m_ayTabAtual, (const void*) ayTab, TAMTAB );
}



/* Define o numero de pecas vivas do time.
**********************************************************/
INT16 CControle::mMRetContaVivas ( INT16 nTime )
{
   INT16
      ncNum = 0;

   // verifica todo conjunto de pecas
   for ( UINT16 ucP = 1 + nTime; ucP <= 12 + nTime; ucP ++ )

      // se peca estah viva conta mais 1
      if ( mRetViva ( ucP ) )
         ncNum ++;

   return ncNum;
}



/* Verifica que tipo de peca estah ocupando uma posicao.
return dc     - dama computador
       pc     - peca comum computador
       v      - posicao desocupada
       pu     - peca comum humano
       du     - dama humano
**********************************************************/
INT16 CControle::mMRetPosTipoPeca ( INT16 nLin, INT16 nCol )
{
   UINT16
      uPeca = mRetPeca ( nLin, nCol );

   // se posicao esta vazia
   if ( uPeca == VAZIO )
      return v;

   // eh dama
   BOOL
      bDama = mRetEhDama ( uPeca );
   // de que time
   BYTE
      yTime = mRetPosTime ( nLin, nCol );

   // se eh dama
   if ( bDama ) {
      // se eh do computador
      if ( yTime == TIMEC )
         return dc;
      else
         // se eh do usuario
         if ( yTime == TIMEU )
            return du;
   }
   else
   // se nao eh dama
      // se eh do computador
      if ( yTime == TIMEC )
         return pc;
      else
         // se eh do usuario
         if ( yTime == TIMEU )
            return pu;
}



/* Faz uma marca no video conforme peca e comando.
**********************************************************/
void CControle::mMMarcaVd ( BYTE yComando, UINT16 uPeca, BYTE yCor )
{
   INT16
      nEsp = 16;    // espacamento entre indicadores de pecas

   CGrVd
      oVd;

   for ( BYTE ycMX = 0; ycMX < 8; ycMX ++ )
      for ( BYTE ycMY = 0; ycMY < nEsp / 2; ycMY ++ ) {
         BYTE
            yCorPonto = yCor + ( ( ycMX + ycMY ) % 2 ) * 5;
         oVd.mAtrVdPixel ( yComando * 10 + ycMX, MAXY - uPeca * nEsp + ycMY, yCorPonto );
      }
}



/* Tenta fazer uma sequencia de movimentos consistente que completa
 uma jogada (come em sequencia e se tem alguma peca que deve
 comer entao joga com essa peca). Transforma peca em dama se for o caso.
Apenas tenta mover a peca selecionada pelo comando definido.
yComando - direcao que deve ser usada no ataque
nTime - time da peca (TIMEC ou TIMEU)
uPSel - valor original da peca que deve ser usada no movimento
return 0 (FALSE) - termina partida;
       1 (TRUE)  - continua partida e moveu com sucesso;
       2         - continua partida, mas nao conseguiu mover peca.
Termina jogo se: nao existem pecas para mover, todas pecas estao bloqueadas.
**********************************************************/
BOOL CControle::mMAutoMove ( BYTE yComando, INT16 nTime, UINT16 uPSel )
{
   // faz testes para avaliar possibilidades de proceguir jogo e movimento
   // verifica se existe alguma jogada possivel
   // salva tabuleiro atual
   BYTE
      ayTabAtual [8][8];
   mMSalvaAtual ( (BYTE*) ayTabAtual );
   // verifica se jogo deve terminar
   UINT16
      uPecaExperimenta = uPSel;
   if ( mAutoMoveObrig ( yComando, nTime, uPecaExperimenta ) == FALSE ) {
      mMRestauraAtual ( (BYTE*) ayTabAtual );
      return FALSE;
   }
   // Verifica se outra peca estah em posicao de comer, se sim nao tenta
   // movimento porque outra peca deve mover (obrigado a comer).
   if ( uPecaExperimenta != uPSel ) {
      mMRestauraAtual ( (BYTE*) ayTabAtual );
      return 2;
   }
   // retaura tabuleiro atual
   mMRestauraAtual ( (BYTE*) ayTabAtual );

   // define direcao de movimentacao para frente conforme time
   INT16
      nDir;
   // se time eh do computador
   if ( nTime == TIMEC )
      nDir = -1;
   else
      // se time eh do usuario
      nDir = 1;

   INT16
      nResMov1,     // resultado do primeiro movimento tentado
      nResMov2;     // resultado dos demais movimentos que ocorreram se
                    // selecionou outra peca ou estah em ataque em sequencia

   // joga com a peca selecionada ou seleciona uma peca que deve comer
   nResMov1 = mMonoMove ( yComando, nDir, uPSel );

   // verifica se ao fazer movimento acima foi feito: nao selecionou outra
   // peca ou nao foi feito por peca estar bloqueada.
   if ( nResMov1 == 0 || nResMov1 == 3 )
      return 2;

   // verifica se o movimento acima formou uma sequencia
   if ( nResMov1 == 2 ) {

      BOOL
         bFimMove = FALSE;
      // enquanto nao terminar sequencia executa
      while ( ! bFimMove ) {

         // tenta atacar para alguma direcao
         nResMov2 = mMonoMove ( 0, nDir, uPSel );
         if ( nResMov2 == 0 )
            nResMov2 = mMonoMove ( 1, nDir, uPSel );
         if ( nResMov2 == 0 )
            nResMov2 = mMonoMove ( 2, nDir, uPSel );
         if ( nResMov2 == 0 )
            nResMov2 = mMonoMove ( 3, nDir, uPSel );

         // verifica se sequencia continua ou se terminou
         if ( nResMov2 != 2 )
            // terminou sequencia
            bFimMove = TRUE;
      }
   }

   // verifica se peca estah em posicao de se tornar dama, se sim transforma
   // verifica se peca jah nao eh dama
   if ( ! mRetEhDama ( uPSel ) )
      // se peca eh do time do usuario, verifica linha 0
      if ( nTime == TIMEU ) {
         // verifica se peca estah na linha extrema
         INT16
            nCol,
            nLin;
         mRetPos ( uPSel, nLin, nCol );
         if ( nLin == 0 )
            // transforma em dama
            mDamifica ( uPSel );
      }
      else
      // se peca eh do time do computador, verifica linha 7
      if ( nTime == TIMEC ) {
         // verifica se peca estah na linha extrema
         INT16
            nCol,
            nLin;
         mRetPos ( uPSel, nLin, nCol );
         if ( nLin == 7 )
            // transforma em dama
            mDamifica ( uPSel );
      }

   return TRUE;
}



/* Funcao de avaliacao estatica.
**********************************************************/

//TABULEIRO = MATRIZ BIDIMENSIONAL DE 0 A 7 (lados:0- computador e 7- usuario
//codificacao:
//pc= peca do computador
//dc= dama do computador
//du= dama do usuario humano
//dc= peca do usuario humano
//npc= numero de pecas do computador
//ndc= numero de damas do computador
//ndu= numero de damas do usuario humano
//ndc= numero de pecas do usuario humano

INT16 CControle::mMAvalia ()
{
   // cria tabuleiro para uso na avaliacao
   BYTE
      ayTab [8][8];    // tabuleiro sem numero individual de cada peca
   // peenche tabuleiro com zero
   memset ( (void*) ayTab, v, TAMTAB );
   // preenche tabuleiro com pecas, faz apenas linhas pares utilizaveis
   for ( INT16 ncL = 0; ncL < 8; ncL += 2 )
      for ( INT16 ncC = 1; ncC < 8; ncC += 2 )
         ayTab [ ncL ] [ ncC ] = mMRetPosTipoPeca ( ncL, ncC );
   // preenche tabuleiro com pecas, faz apenas linhas impares utilizaveis
   for ( INT16 ncL = 1; ncL < 8; ncL += 2 )
      for ( INT16 ncC = 0; ncC < 8; ncC += 2 )
         ayTab [ ncL ] [ ncC ] = mMRetPosTipoPeca ( ncL, ncC );

//** PESOS **//

//#define pesoavancopeca 3
//#define pesoataquecostas 7
INT16 /*unsigned int*/ nota;
int inter=0;
nota=1;
int npc=0;
int npu=0;
int ndc=0;
int ndu=0;
unsigned int MASTER=0;
unsigned int avanco=0;
int pontos=0;

// 
// 
//         SUPER OTIMIZACAO DA FUNCAO DE AVALIACAO           
// 
// 

for ( int Lin = 0; Lin < 8; Lin ++ )
   for ( int Col = 0; Col < 8; Col ++ ){


       // CONTAGEM DO NUMERO DE PECAS DO JOGO
       // PARA POSSIBILITAR CALCULO DA ESTRATEGIA
       // PRINCIPAL (MASTER)
       //        ||
       //        ||
       //        \/
       if (ayTab [ Lin ] [ Col ] == pc)
         npc ++;
       else{
       if (ayTab [ Lin ] [ Col ] == pu)
         npu ++;
       else{
       if (ayTab [ Lin ] [ Col ] == dc)
            ndc ++;
       else
       if (ayTab [ Lin ] [ Col ] == du)
             ndu ++;}}


       // CALCULO DA ESTRATEGIA DE AVANCO EXPONENCIAL
       // DE PECAS DO COMPUTADOR (n^2)
       // UNIDADES DE PONDERACAO = 2 NA NUMERO DA LINHA (0 A 7)
       //        ||
       //        ||
       //        \/
       if (ayTab [ Lin ] [ Col ] == pc)
         avanco=avanco+Lin*Lin ;


       // CALCULO DA ESTRATEGIA DE ATAQUE DA DAMA POR TRAS `A PECAS
       // DO COMPUTADOR E ESTRATEGIA DE COMBATE DE DAMAS
       // 1 CASA  ATRAS DA PECA -> 40 UNIDADES DE PONDERACAO
       // 2 CASAS ATRAS DA PECA -> 30 UNIDADES DE PONDERACAO
       // 3 CASAS ATRAS DA PECA -> 20 UNIDADES DE PONDERACAO
       // 4 CASAS ATRAS DA PECA -> 10 UNIDADES DE PONDERACAO
       // 5 CASAS ATRAS DA PECA -> 05 UNIDADES DE PONDERACAO
       //        ||
       //        ||
       //        \/
       if (ayTab [ Lin ] [ Col ] == pu){//begin if3
         if (ayTab [(Lin+1)] [Col-1] == dc)
         pontos=pontos+60;
         else{
         if (ayTab [(Lin+1)] [Col+1] == dc)
         pontos=pontos+60;
         else{
         if (ayTab [(Lin+2)] [Col-2] == dc)
         pontos=pontos+30;
         else{
         if (ayTab [(Lin+2)] [Col+2] == dc)
         pontos=pontos+30;
         else{
         if (ayTab [(Lin+3)] [Col-3] == dc)
         pontos=pontos+20;
         else{
         if (ayTab [(Lin+3)] [Col+3] == dc)
         pontos=pontos+20;
         else{
         if (ayTab [(Lin+4)] [Col-4] == dc)
         pontos=pontos+10;
         else{
         if (ayTab [(Lin+4)] [Col+4] == dc)
         pontos=pontos+10;
         if (ayTab [(Lin+5)] [Col-5] == dc)
         pontos=pontos+5;
         else
         if (ayTab [(Lin+5)] [Col+5] == dc)
         pontos=pontos+5;}}}}}}}//end elses
         }//end if3
       if (ayTab [ Lin ] [ Col ] == du){//begin if3
         if (ayTab [(Lin+1)] [Col-1] == dc)
         pontos=pontos+10;
         else{
         if (ayTab [(Lin+1)] [Col+1] == dc)
         pontos=pontos+10;
         else{
         if (ayTab [(Lin+2)] [Col-2] == dc)
         pontos=pontos+8;
         else{
         if (ayTab [(Lin+2)] [Col+2] == dc)
         pontos=pontos+8;
         else{
         if (ayTab [(Lin+3)] [Col-3] == dc)
         pontos=pontos+6;
         else{
         if (ayTab [(Lin+3)] [Col+3] == dc)
         pontos=pontos+6;
         else{
         if (ayTab [(Lin+4)] [Col-4] == dc)
         pontos=pontos+4;
         else{
         if (ayTab [(Lin+4)] [Col+4] == dc)
         pontos=pontos+4;
         if (ayTab [(Lin+5)] [Col-5] == dc)
         pontos=pontos+2;
         else
         if (ayTab [(Lin+5)] [Col+5] == dc)
         pontos=pontos+2;}}}}}}}//end elses
         }//end if3
       if (ayTab [ Lin ] [ Col ] == du){//begin if3
         if (ayTab [(Lin-1)] [Col-1] == dc)
         pontos=pontos+10;
         else{
         if (ayTab [(Lin-1)] [Col+1] == dc)
         pontos=pontos+10;
         else{
         if (ayTab [(Lin-2)] [Col-2] == dc)
         pontos=pontos+8;
         else{
         if (ayTab [(Lin-2)] [Col+2] == dc)
         pontos=pontos+8;
         else{
         if (ayTab [(Lin-3)] [Col-3] == dc)
         pontos=pontos+6;
         else{
         if (ayTab [(Lin-3)] [Col+3] == dc)
         pontos=pontos+6;
         else{
         if (ayTab [(Lin-4)] [Col-4] == dc)
         pontos=pontos+4;
         else{
         if (ayTab [(Lin-4)] [Col+4] == dc)
         pontos=pontos+4;
         if (ayTab [(Lin-5)] [Col-5] == dc)
         pontos=pontos+2;
         else
         if (ayTab [(Lin-5)] [Col+5] == dc)
         pontos=pontos+2;}}}}}}}//end elses
         }//end if3

      }//endfor


   // 
   //  PONDERACAO DE ESTRATEGIAS 
   // 
   // MASTER =((7*pontos)+(3*avanco))/10
   // INTER  =(7*pontos)+(3*avanco))/10
   // NOTA   = MASTER + INTER -32000




//CALCULO DA MEDIA DE PRIORIDADES DE ESTRATEGIA INTERMEDIARIAS
// PONTOS (PESO 7)=ESTRATEGIA DE ATAQUE POR TRAS A PECAS DO ADVERSARIO
//                 PELA DAMA DO COMPUTADOR E COMBATE DE DAMAS
// AVANCO (PESO 3)=ESTRATEGIA DE AVANCO ESPONENCIAL (n^2) DAS PECAS DO
//                 COMPUTADOR EM BUSCA DA TRANSFORMACAO EM DAMA
//        ||
//        ||
//        \/
inter=((5*pontos)+(5*avanco))/10;


//PARA GARANTIR QUE NENHUMA ESTRATEGIA DE ATAQUE COMO ATAQUE POR TRAZ OU
//AVANCO EXPONECIAL
//TENHA MAIS OU MENOS PRIORIDADE QUE CHANCES DE COMER (+1000 PRIORITY)
//OU RISCO DE PERDAS DE PECA (-1000 PRIORITY)
//        ||
//        ||
//        \/
if (inter>950)
   inter=950;


// CALCULO DA PRIORIODADE DE ESTRATEGIA PRINCIPAL - MASTER
// MASTER= MAIOR DIFERENCA FAVORAVEL POSSIVEL DE PECAS DO COMPUTADOR
// MENOS PECAS DO ADVERSARIO (DAMA TEM PESO 3 NO CALCULO)
// PECAS COMUNS -> 1000 UNIDADES DE PONDERACAO
// DAMAS        -> 2000 UNIDADES DE PONDERACAO
//        ||
//        ||
//        \/
MASTER=((npc-npu)+13+(ndc*2-ndu*2)+37)*1000;


//CALCULO DA NOTA FINAL -32000 QUE EH UM DESLOCAMENTO PARA GARANTIR
//QUE O RESULTADO CAIBA EM VARIAVEIS "INT" (~-32000 A ~+32000)
//ISSO EH FEITO PARA GANHO DE VELOCIDADE DE PROCESSAMENTO
//        ||
//        ||
//        \/
nota =(MASTER) +(inter)-32000;

return nota;
}



/* Executa minimax ateh atingir a profundidade maxima definida.
nCmd - comando que peca deve executar;
uPeca - peca selecionada para movimento;
nMaxProf - numero maximo de niveis (profundidade);
nNivel - nivel atual na arvore de possibilidades.
return - peso (nota) do nodo.
**********************************************************/
INT CControle::mMinimax ( INT16 nCmd, UINT16 uPeca, INT16 nMaxProf,
                          INT16 nNivel )
{
   // salva tabuleiro atual
   BYTE
      ayTabAtual [8][8];
   mMSalvaAtual ( (BYTE*) ayTabAtual );

   // faz jogada hipotetica alterando tabuleiro atual da classe
   // verifica quem estah jogando: nivel impar eh computador
   BYTE
      nResMov;
   if ( nNivel % 2 == 1 ) {

      // faz jogada
      nResMov = mMAutoMove ( nCmd, TIMEC, uPeca );

      // verifica se peca estava bloqueada, se sim termina com pior valor
      if ( nResMov == 2 ) {
         // retaura tabuleiro atual
         mMRestauraAtual ( (BYTE*) ayTabAtual );
         // retorna pior valor para computador numa maximizacao
         return - 1000000;
      }
   }
   else {

      // faz jogada
      nResMov = mMAutoMove ( nCmd, TIMEU, uPeca );

      // verifica se peca estava bloqueada, se sim termina com pior valor
      if ( nResMov == 2 ) {
         // retaura tabuleiro atual
         mMRestauraAtual ( (BYTE*) ayTabAtual );
         // retorna pior valor para computador numa minimizacao
         return 1000000;
      }
   }

   // se atingiu profundidade suficiente avalia aquele tabuleiro final
   INT
      nPesoNodo;
   if ( nNivel == nMaxProf )

      // avalia jogada hipotetica
      nPesoNodo = mMAvalia ();

   else
   // se profundidade insuficiente gera outro conjunto de jogadas

      // Verifica quem estah jogando: usuario (minimizacao) ou computador
      // (maximizacao), Usado para definir jogadas de quem prever.
      // Se for nivel impar (jogada do computador) preve proximo nivel que eh
      // minimizacao - usuario.
      if ( nNivel % 2 == 1 ) {

         INT
            nValorNodoHip = 1000000;      // terah o menor valor possivel

         // gera cada possibilidade de jogo do usuario (proximo a jogar)
         // para cada possivel peca
         for ( UINT16 ucP = 1 + TIMEU; ucP <= 12 + TIMEU; ucP ++ ) {

            // define limite de possibilidades de comandos para peca
            INT16
               nLimiteCmd = 1;
            // verifica se peca eh dama, se sim gera comandos para damas tambem
            if ( mRetEhDama ( ucP ) )
               nLimiteCmd = 3;

            // para cada possivel direcao de peca
            for ( INT16 ncCmd = 0; ncCmd <= nLimiteCmd; ncCmd ++ )

               // se peca estah viva
               if ( mRetViva ( ucP ) ) {

                  // gera um ramo do minimax para cada peca e comando
                  INT
                     nValorRamo = mMinimax ( ncCmd, ucP, nMaxProf, nNivel + 1 );

                  // verifica se valor do ramo gerado eh menor que atual
                  if ( nValorRamo < nValorNodoHip )
                     // atualiza com menor valor (minimiza)
                     nValorNodoHip = nValorRamo;
               }
         }

         // define valor desse nodo como menor valor
         nPesoNodo = nValorNodoHip;
      }
      else {
      // se eh nivel par serah maximizacao, proximo a jogar computador

         INT
            nValorNodoHip = - 1000000;      // terah o maior valor possivel

         // gera cada possibilidade de jogo do computador (proximo a jogar)
         // para cada possivel peca
         for ( UINT16 ucP = 1 + TIMEC; ucP <= 12 + TIMEC; ucP ++ ) {

            // define limite de possibilidades de comandos para peca
            INT16
               nLimiteCmd = 1;
            // verifica se peca eh dama, se sim gera comandos para damas tambem
            if ( mRetEhDama ( ucP ) )
               nLimiteCmd = 3;

            // para cada possivel direcao de peca
            for ( INT16 ncCmd = 0; ncCmd <= nLimiteCmd; ncCmd ++ )

               // se peca estah viva
               if ( mRetViva ( ucP ) ) {

                  // gera um ramo do minimax para cada peca e comando
                  INT
                     nValorRamo = mMinimax ( ncCmd, ucP, nMaxProf, nNivel + 1 );

                  // verifica se valor do ramo gerado eh maior que atual
                  if ( nValorRamo > nValorNodoHip )
                     // atualiza com maior valor (maximiza)
                     nValorNodoHip = nValorRamo;
               }
         }

         // define valor desse nodo como maior valor
         nPesoNodo = nValorNodoHip;
      }

   // restaura tabuleiro atual
   mMRestauraAtual ( (BYTE*) ayTabAtual );

   return nPesoNodo;
}
/* minimax de 1 nivel
{
   // salva tabuleiro atual
   BYTE
      ayTabAtual [8][8];
   mMSalvaAtual ( (BYTE*) ayTabAtual );

   // faz jogada hipotetica alterando tabuleiro atual da classe
   mAutoMoveObrig ( nCmd, TIMEC, uPeca );

   // avalia jogada hipotetica
   INT
      nPesoNodo = mMAvalia ();

   // retaura tabuleiro atual
   mMRestauraAtual ( (BYTE*) ayTabAtual );

   return nPesoNodo;
}*/



/* Faz jogada do computador alterando tabela de posicao de pecas
nMinNiveis - numero minimo de niveis a serem explorados pelo minimax.
return 0 (FALSE) - termina partida;
       1 (TRUE)  - continua partida.
**********************************************************/
BOOL CControle::mJogadaComputador ( INT16 nMinNiveis )
{
   m_yUltProf = 0;

   // verifica se numero de niveis eh zero, se for faz jogadas aleatorias
   if ( nMinNiveis == 0 ) {

         // valor que serah usado para selecionar primeira peca que tentarah mover
      UINT16
         uPSel = ( random () % NPECAS );   // numero de 0 a NPECAS - 1

      // define comando: direcao de ataque
      BYTE
         yComando = ( random () % 4 );

      BOOL
         bResMov = mAutoMoveObrig ( yComando, TIMEC, uPSel );

      m_yUltPeca = uPSel;
      m_yUltComando = yComando;

      return bResMov;
   }

   // se deve executar minimax

   UINT16
      uMPeca = 1;  // melhor peca para mover
   INT16
      nMCmd = 0;   // melhor comando a ser executado pela melhor peca
   INT
      nMPeso = - 1000000; // peso da melhor jogada, inicia por maximizacao

   // define profundidade maxima do minimax, conforme numero de pecas
   INT16
      nProf = nMinNiveis;   // originalmente serah a quantidade default
   // verifica se existem 6 ou menos pecas no tabuleiro, se sim mais 1 nivel
   /*if ( mMRetContaVivas ( TIMEC ) + mMRetContaVivas ( TIMEU ) <= 6 )
      nProf ++;*/

   // gera cada possibilidade de jogo
   // para cada possivel peca
   for ( UINT16 ucP = 1; ucP <= 12; ucP ++ ) {

      // define limite de possibilidades de comandos para peca
      INT16
         nLimiteCmd = 1;
      // verifica se peca eh dama, se sim gera comandos para damas tambem
      if ( mRetEhDama ( ucP ) )
         nLimiteCmd = 3;

      // para cada possivel direcao de peca
      for ( INT16 ncCmd = 0; ncCmd <= nLimiteCmd; ncCmd ++ ) {

         // faz marca no video de inicio de pesquisa do ramo
         mMMarcaVd ( ncCmd, ucP, 250 );

         // se peca estah viva
         if ( mRetViva ( ucP ) ) {

            // gera um ramo do minimax para cada peca e comando
            INT
               nPesoRamo = mMinimax ( ncCmd, ucP, nProf, 1 );

            // Se o resultado foi maior que o atual guarda para uso futuro.
            // Se for igual ao atual sorteia qual dos dois vai usar.
            if ( nPesoRamo > nMPeso ||
                 ( nPesoRamo == nMPeso && ( rand () % 2 ) ) ) {

               // guarda comando, peca e atualiza melhor peso
               uMPeca = ucP;
               nMCmd = ncCmd;
               nMPeso = nPesoRamo;
            }

            // faz marca no video de fim de pesquisa do ramo, se peca viva
            mMMarcaVd ( ncCmd, ucP, 240 );
         }
         else
            // faz marca no video se peca morta
            mMMarcaVd ( ncCmd, ucP, 230 );

      }
   }

   // ultimo movimento do computador com minimax
   m_yUltProf = nProf;
   m_yUltPeca = uMPeca;
   m_yUltComando = nMCmd;

   // faz jogada que recebeu maior peso
   return mAutoMoveObrig ( nMCmd, TIMEC, uMPeca );
}



/* Testa se o usuario fez uma jogada consistente, atualiza
 tabelas de acordo. yComando pode ser:
 T_HOME, T_PG_UP, T_END, T_PG_DN que serao mapeados para
 CMD_HOME == 0, CMD_PG_UP == 1, CMD_END == 2, CMD_PG_DN == 3.
 uPeca informa peca selecionada do tabuleiro virtual (possui
 o valor original da peca no jogo, nao valor atual que eh
 dependente da peca ser dama).
 return 0 - jogada invalida ou deve continuar (se comer em serie);
        1 - jogada terminada;
        2 - terminou jogo
**********************************************************/
BOOL CControle::mJogadaUsuario ( BYTE yComando, UINT16& uPeca )
{
   BOOL
      bResMov = TRUE;    // inicialmente supoe que jogada vai terminar

   MAPEIA_T_CMD;

   // faz movimentacao e/ou ataque das pecas
   bResMov = mMonoMove ( yComando, 1, uPeca );

   // converte resposta: 0, 2, 3 sao FALSE, 1 TRUE
   if ( bResMov != 1 )
      bResMov = FALSE;

   // verifica se deve tornar peca dama
   // verifica se peca jah nao eh dama
   if ( ! mRetEhDama ( uPeca ) ) {
      // verifica se peca estah na linha extrema
      INT16
         nCol,
         nLin;
      mRetPos ( uPeca, nLin, nCol );
      if ( nLin == 0 )
         // transforma em dama
         mDamifica ( uPeca );
   }

   return bResMov;
}
