/* Altera dados do solido: ordenacao de poligonos, cor de poligonos, ... */



#include <stdio.h>
#include <string.h>

#include "defs.h"

#include "solido.h"

#include "arquivo.h"

#include "arq3d.h"

#include "universo.h"

#include "sr.h"

#include "transf.h"

#include "contrsol.h"



/* construtor - inicia sem objeto definido
**********************************************************/
CControleSolido::CControleSolido ()
{
   m_poS = NULL;
}



/* construtor - inicia jah definindo solido a ser manipulado
**********************************************************/
CControleSolido::CControleSolido ( CSolido& oS )
{
   m_poS = &oS;
}



/* Define solido a ser manipulado
**********************************************************/
void CControleSolido::mAtrObjeto ( CSolido& oS )
{
   m_poS = &oS;
}




/* Define poligonos que deverao ser ordenados. Chamado uma
vez no solido temporario antes de iniciar ordenacoes de poligonos.
**********************************************************/
void CControleSolido::mDefOrdenar ()
{
   // calcula distancia do ponto medio dos poligonos ateh a origem
   for ( UINT16 ucP = 0; ucP < m_poS->m_uNumPoligonos; ucP ++ )

      // guarda indice
      * ( m_poS->m_pauP + ucP ) = ucP;
}



/* Atualiza matriz de distancias de poligonos e faz ordenacao
dos poligonos conforme a distancia em Z. Para definir o poligono
cuja distancia serah calculada eh usado o indice que estah na tabela
de indices de poligonos ordenados por maior distancia. Apohs ser
calculada a distancia em Z de cada poligono eh feita a atualizacao da
tabela de indices de poligonos ordenados, reordenando-os conforme as
novas distancias obtidas. A ordenacao eh decrescente. Usa bubblesort.
O efeito dessa forma de ordenacao eh que apenas os poligonos que
trocaram sua proximidade com a origem em relacao a um outro poligono
eh que terao a sua ordem alterada (se um poligono que estava mais a
frente que outro passou para tras entao ele terah sua ordem alterada).
**********************************************************/
void CControleSolido::mOrdena ()
{
   // calcula distancia do ponto medio do poligono ateh a origem
   for ( UINT16 ucP = 0; ucP < m_poS->m_uNumPoligonos; ucP ++ ) {

      UINT16
         uNdxProx = * ( m_poS->m_pauP + ucP );

      // calcula distancia Z do ponto medio
      * ( m_poS->m_panD + ucP ) = (INT16)
              ( * ( m_poS->m_panZ + * ( m_poS->m_pauV1 + uNdxProx ) ) +
                * ( m_poS->m_panZ + * ( m_poS->m_pauV2 + uNdxProx ) ) +
                * ( m_poS->m_panZ + * ( m_poS->m_pauV3 + uNdxProx ) ) +
                * ( m_poS->m_panZ + * ( m_poS->m_pauV4 + uNdxProx ) ) ) >> 2;
   }

   // ordena indices conforme proximidade com a origem, usa bubblesort
   BOOL
      bFim;
   UINT16
      uPosFim = m_poS->m_uNumPoligonos;
   // faz ateh estar ordenado (nao haverem trocas por uma varredura)
   do {

      bFim = TRUE;

      // executa cada interacao varrento a matriz de distancias
      for ( UINT16 ucPos = 0; ucPos < ( uPosFim - 1 ); ucPos ++ )

         // se uma distancia eh menor que a seguinte entao troca
         if ( * ( m_poS->m_panD + ucPos ) < * ( m_poS->m_panD + ucPos + 1 ) ) {

            INT16
               nDist = * ( m_poS->m_panD + ucPos );
            UINT16
               uNdx  = * ( m_poS->m_pauP + ucPos );
            * ( m_poS->m_panD + ucPos ) = * ( m_poS->m_panD + ucPos + 1 );
            * ( m_poS->m_pauP + ucPos ) = * ( m_poS->m_pauP + ucPos + 1 );
            * ( m_poS->m_panD + ucPos + 1 ) = nDist;
            * ( m_poS->m_pauP + ucPos + 1 ) = uNdx;
            bFim = FALSE;
         }

      uPosFim --;

   } while ( ! bFim );
}



/* Muda cor de um conjunto de poligonos
**********************************************************/
void CControleSolido::mAtrCorPols ( UINT16 uPolIni, UINT16 uNumPols, BYTE yCor )
{
   memset ( (void*) ( m_poS->m_payCor + uPolIni ), yCor, uNumPols );
}
