/* Implementa funcoes de handler de teclado para deteccao de varias teclas
pressionadas simultaneamente. Eh usado para substituir a funcao de
tratamento do teclado da BIOS (handler de teclado da BIOS) por uma outra
funcao fornecida pelo programa. Essa funcao do programa nao usa buffer de
teclado e ignora os tempos de repeticao automatica de caracteres.

TODAS AS TECLAS DEVEM ESTAR SOLTAS QUANDO O HANDLER FOR REMOVIDO, SENAO GPF
NO WINDOWS!!!

Numero da interrupcao do teclado: 9
Porta para leitura de teclas: 0x60
Porta para aviso de recebimento de interrupcao: 0x20, valor 0x20
Codigo dos bytes enviados para o handler (nao tratado PrtScr e Ctrl+Break):
 Se for tecla especial:
   valor 224
   Byte seguinte (na proxima chamada ao handler):
   Se for tecla sendo pressionada
     valor codigo_tecla        (posicao na tabela: codigo_tecla + 128)
   Se for tecla sendo solta
     valor codigo_tecla + 128  (posicao na tabela: codigo_tecla + 128)
 Se nao for tecla especial
   Se for tecla sendo pressionada
     valor codigo_tecla
   Se for tecla sendo solta
     valor codigo_tecla + 128                                              */



#include <stdio.h>
#include <dpmi.h>
#include <go32.h>
#include <string.h>
#include <dos.h>

#include "../defs.h"

// Devem ser globais para handler acessar. Se nao forem globais alteracoes
// feitas nao sao visiveis para demais funcoes locais.
BOOL ayTECLAS [ 256 ];    // indica tecla pressionada, TRUE, ou nao, FALSE
BOOL bESPECIAL = FALSE;   // indica tecla especial sendo tratada
BOOL bINSTALADO = FALSE;  // Indica que handler estah ou nao instalado, evita
                          // instalar ou desinstalar handler mais de uma vez.
_go32_dpmi_seginfo m_stHANDLER_ANTIGO;    // handler de teclado original
_go32_dpmi_seginfo m_stHANDLER_NOVO;      // handler de teclado do programa

#include "teclado.h"



/* construtor - Se bInstalarHandler = TRUE substitui a
funcao de tratamento do teclado atual pela funcao do programa.
Default bInstalarHandler = FALSE
**********************************************************/
CTeclado::CTeclado ( BOOL bInstalarHandler )
{
   m_bInstalarHandler = bInstalarHandler;

   // verifica se deve instalar handler
   if ( m_bInstalarHandler )
      mInstalaHandler ();
}



/* destrutor - substitui o handler do programa pelo original.
Agaurda ateh todas as teclas estarem soltas para isso.
**********************************************************/
CTeclado::~CTeclado ()
{
   // verifica se deve desinstalar handler
   if ( m_bInstalarHandler )
      mDesinstalaHandler ();
}



/* Handler de telcado (captura teclas), chamado cada vez que
uma tecla eh pressionada ou solta, ou seja, sempre que eh
gerada a interrupcao do teclado (numero 9).
VER mFimHandler___!!!!!!!!!
**********************************************************/
void CTeclado::mHandlerNovo ()
{
   unsigned char
      yTecla = inportb ( 0x60 );

   // verifica se nao foi pressionado anteriormente um tecla especial
   if ( ! bESPECIAL )
      // detecta pressionamento de tecla especial
      if ( yTecla == 224 )
         bESPECIAL = TRUE;
      else
         // trata tecla nao especial
         // altera posicao na tabela correspondente a tecla pressionada
         // se tecla estah sendo pressionada (bit 7 == 0) altera posicao
         // correspondente para TRUE senao para FALSE
         if ( ( yTecla & 128 ) == 0 )
            * ( ayTECLAS + ( yTecla & 127 ) ) = TRUE;
         else
            // senao altera para FALSE
            * ( ayTECLAS + ( yTecla & 127 ) ) = FALSE;
   else {
      // trata tecla especial
      // se foi pressionada
      if ( ( yTecla & 128 ) == 0 )
         * ( ayTECLAS + 128 + ( yTecla & 127 ) ) = TRUE;
      else
         // se foi solta
         * ( ayTECLAS + 128 + ( yTecla & 127 ) ) = FALSE;

      bESPECIAL = FALSE;
   }


   // confirma recebimento da interrupcao
   outportb ( 0x20, 0x20 );
}



/* Usado apenas como referencia para saber tamanho do
codigo executavel da funcao mHandlerNovo: (mHandlerNovo___ -
- mHandlerNovo). DEVE VIR SEMPRE LOGO APOS mHandlerNovo!!!
**********************************************************/
void CTeclado::mHandlerNovo___ ()
{
}



/* Instala o handler de teclado. Pode ser usado para reinstalar
o handler se foi temporariamente removido por mDesinstalaHandler.
**********************************************************/
void CTeclado::mInstalaHandler ()
{
   // verifica se handler nao estah instalado
   if ( ! bINSTALADO ) {
      // antes de substituir pelo novo handler eh necessario salvar o antigo
      _go32_dpmi_get_protected_mode_interrupt_vector ( 9, &m_stHANDLER_ANTIGO );

      // reserva memoria para as funcoes do handler e chamadas pelo handler
      //  (impossibilita swap como memoria virtual)
      _go32_dpmi_lock_code ( (void*) mHandlerNovo, (UINT) mHandlerNovo___ - (UINT) mHandlerNovo );

      // reserva memoria do handler (impossibilita swap como memoria virtual)
      _go32_dpmi_lock_data ( ayTECLAS, sizeof ( ayTECLAS ) );

      // substitui o handler de teclado atual pelo fornecido pelo programa
      m_stHANDLER_NOVO.pm_offset = (INT) mHandlerNovo;
      m_stHANDLER_NOVO.pm_selector = _go32_my_cs();
      _go32_dpmi_allocate_iret_wrapper ( &m_stHANDLER_NOVO );
      _go32_dpmi_set_protected_mode_interrupt_vector ( 9, &m_stHANDLER_NOVO );

      // limpa a tabela de teclas pressionadas
      memset ( (void*) ayTECLAS, FALSE, sizeof ( ayTECLAS ) );

      // define handler intalado
      bINSTALADO = TRUE;
   }
}



/* Instala o handler de teclado. Pode ser usado para desinstalar
temporariamente o handler de teclado e em seguida reinstalar
chamando mInstalaHandler.
**********************************************************/
void CTeclado::mDesinstalaHandler ()
{
   // verifica se handler estah instalado
   if ( bINSTALADO ) {

      // todas teclas devem estar soltas serao pode causar GPF
      mRetSoltas ();

      // substitui o handler de teclado atual pelo antigo
      _go32_dpmi_set_protected_mode_interrupt_vector ( 9, &m_stHANDLER_ANTIGO );

      // libera memoria do handler do programa
      _go32_dpmi_free_iret_wrapper ( &m_stHANDLER_NOVO );

      // define handler desinstalado
      bINSTALADO = FALSE;
   }
}



/* Verifica se uma determinada tecla estah pressionada
 no momento. return TRUE - sim; FALSE - nao.
**********************************************************/
BOOL CTeclado::mRetPress ( BYTE yTecla )
{
   return * ( ayTECLAS + yTecla );
}



/* Retorna codigo de uma tecla pressionada no momento ou 0
se nenhuma tecla estiver pressionada (nao testa tecla 0, se
eh que existe uma tecla com esse codigo).
**********************************************************/
BYTE CTeclado::mRetPress ()
{
   // varre todo vetor de teclas
   for ( UINT16 ucT = 0; ucT < 256; ucT ++ )
      // se encontrou alguma tecla pressionada (TRUE) retorn seu codigo
      if ( mRetPress ( ucT ) == TRUE )
         return ucT;

   return 0;
}



/* Aguarda ateh que alguma tecla seja pressionada. return
codigo da tecla pressionada. Avalia cada tecla do vetor de
teclas em sequencia ateh o seu fim e entao volta ao inicio
e repete. Se bPress == TRUE (default) somente comeca a testar teclas
apos todas estarem soltas, se for FALSE e uma tecla jah estava
pressionada entao fica no loop apenas 1 volta e pode confundir
usuario por ser muito rapido.
**********************************************************/
BYTE CTeclado::mRetTeclas ( BOOL bTestaPress )
{
   // Se todas as teclas devem estar soltas antes de aguardar que alguma
   // tecla seja pressionada.
   if ( bTestaPress ) {
      BOOL
         bSoltas;
      // enquanto uma tecla estiver pressionada
      do {
         bSoltas = TRUE;
         // varre todo vetor de teclas
         for ( UINT16 ucT = 0; ucT < 256; ucT ++ )
            // se encontrou alguma tecla pressionada (TRUE)
            if ( mRetPress ( ucT ) == TRUE )
               bSoltas = FALSE;
      } while ( bSoltas == FALSE );
   }

   // enquanto uma tecla nao for pressionada
   while ( 1 )
      // varre todo vetor de teclas
      for ( UINT16 ucT = 0; ucT < 256; ucT ++ )
         // se encontrou alguma tecla pressionada (TRUE) retorn seu codigo
         if ( mRetPress ( ucT ) == TRUE )
            return ucT;
}



/* Aguarda ateh que a tecla yTecla seja pressionada.
Se bPress = TRUE somente comeca a testar a tecla
apos essa tecla estar solta, se for FALSE e a tecla jah estava
pressionada entao fica no loop apenas 1 volta e pode confundir
usuario por ser muito rapido.
**********************************************************/
void CTeclado::mRetTecla ( BYTE yTecla, BOOL bTestaPress )
{
   // se eh para testar se tecla estava pressionada antes
   if ( bTestaPress )
      // enquanto a tecla estiver pressionada
      while ( mRetPress ( yTecla ) == TRUE );

   // enquanto a tecla nao for pressionada
   while ( mRetPress ( yTecla ) == FALSE );
}



/* Aguarda ateh todas as teclas estarem soltas
**********************************************************/
void CTeclado::mRetSoltas ()
{
   BOOL
      bSoltas;
   // enquanto uma tecla estiver pressionada
   do {
      bSoltas = TRUE;
      // varre todo vetor de teclas
      for ( UINT16 ucT = 0; ucT < 256; ucT ++ )
         // se encontrou alguma tecla pressionada (TRUE)
         if ( mRetPress ( ucT ) == TRUE )
            bSoltas = FALSE;
   } while ( bSoltas == FALSE );
}
