/* Funcoes graficas basicas com memoria virtual de video.

Assembler AT&T + DJGPP:

- se gravar em variavel de memoria passada pelos parametros
  por "memory" no 3 :;
- __volatile__ permite ao compilador reorganizar o codigo */



#include <string.h>   // necessario para acessar sys/movedata.h
#include <stdio.h>
#include <stdlib.h>

#ifdef TRG_DOS_DJGPP
#include <dos.h>
#include <go32.h>
#include <conio.h>
#include <pc.h>
#endif

#ifdef TRG_LINUX_SVGA
#include <vga.h>
#endif

#include "defs.h"

#include "grvd.h"

#include "gr.h"



// calcula incremento de x (inverso da declividade)
#define INCREMENTOX( XI, YI, XF, YF ) ( (float) ( ( XF ) - ( XI ) ) / ( ( YF ) - ( YI ) ) )



/* construtor - Aloca memoria.
bInicGr == TRUE (default) aloca memoria virtual de video.
Altera comportamento do destrutor.
**********************************************************/
//CGr::CGr ( BOOL bInicGr )



/* destrutor - Volta ao modo texto, desaloca memoria
**********************************************************/
CGr::~CGr ()
{
   // verifica se entrou em modo grafico no inicio da classe
   if ( m_bInicGrVd == TRUE ) {

      // desaloca
      delete m_payVr [ 0 ];
      if ( m_payVr [ 1 ] != NULL )
         delete m_payVr [ 1 ];
   }
}



/* Espera retraco de video para desenhar (=="WaitRetrace"),
  durante a espera zera a pagina de video virtual nao ativa.
__asm__ __volatile__ ("

      movl  %0,      %%edi   edi com endereco inicial a ser zerado
      movl  $16000,  %%ecx   ecx recebe o numero de interacoes (16000)
      xorl  %%ebx,   %%ebx   ebx recebe 0 (valor a ser escrito)

      movw  $0x03DA, %%dx    parte do "WaintRetrace"

  A1: movl  %%ebx,  (%%edi)  escreve no endereco apontado por edi um zero de 4 bytes (conteudo de ebx)
      addl  $4,      %%edi   soma 4 bytes no edi, proximos 4 bytes a serem escritos
      subl  $1,      %%ecx   subtrai 1 de ecx, menos uma interacao
      cmpl  $0,      %%ecx   se nao existe mais memoria para zerar contador de interacoes (ecx) chegou a zero nao zera mais a memoria
      jz    A2               se ecx == 0 desvia para parte onde soh faz primeiro teste do "WaitRetrace"

      inb   %%dx,    %%al    primeiro teste do "WaitRetrace"
      andb  $0x8,    %%al
      jz    A1               se o teste nao teve sucesso volta

  A4: movl  %%ebx,  (%%edi)  mesma escrita em endereco que em A1, soh depois de passar primeiro teste do "WaitRetrace"
      addl  $4,      %%edi
      subl  $1,      %%ecx
      cmpl  $0,      %%ecx
      jz    A3               se terminou de zerar memoria desvia para parte sem escrita em memoria (soh teste)

      inb   %%dx,    %%al    segundo teste do "WaitRetrace"
      andb  $0x8,    %%al
      jz    A4               se falhou teste volta

      jmp   AFimAtr          se acertou teste termina escrita na memoria antes de sair

  A2: inb   %%dx,    %%al    primeiro teste do "WaitRetrace" sem zerar memoria
      andb  $0x8,    %%al    (igual a "WaitRetrace" normal)
      jnz   A2

  A3: inb   %%dx,    %%al    segundo teste do "WaitRetrace" sem zerar memoria
      andb  $0x8,    %%al    (igual a "WaitRetrace" normal)
      jz    A3

      jmp AFim               termina sem passar por termino de escrita em memoria (logo abaixo), escrita jah acabou

  AFimAtr:
      movl  %%ebx,  (%%edi)  termina de escrever a memoria (apohs ter passado pelos 2 testes de "WaitRetrace")
      addl  $4,      %%edi
      subl  $1,      %%ecx
      cmpl  $0,      %%ecx
      jnz   AFimAtr

  AFim:

**********************************************************/
void CGr::mVrSincRetraco ()
{
   __asm__ __volatile__ ("

          movl  %0,      %%edi
          movl  $16000,  %%ecx
          xorl  %%ebx,   %%ebx

          movw  $0x03DA, %%dx

      A1: movl  %%ebx,  (%%edi)
          addl  $4,      %%edi
          subl  $1,      %%ecx
          cmpl  $0,      %%ecx
          jz    A2

          inb   %%dx,    %%al
          andb  $0x8,    %%al
          jz    A1

      A4: movl  %%ebx,  (%%edi)
          addl  $4,      %%edi
          subl  $1,      %%ecx
          cmpl  $0,      %%ecx
          jz    A3

          inb   %%dx,    %%al
          andb  $0x8,    %%al
          jz    A4

          jmp   AFimAtr

      A2: inb   %%dx,    %%al
          andb  $0x8,    %%al
          jnz   A2

      A3: inb   %%dx,    %%al
          andb  $0x8,    %%al
          jz    A3

          jmp   AFim

      AFimAtr:
          movl  %%ebx,  (%%edi)
          addl  $4,      %%edi
          subl  $1,      %%ecx
          cmpl  $0,      %%ecx
          jnz   AFimAtr

      AFim:
      "
      :
      : "g" ( m_payVr [ ! m_nVrAtiva ] )
      : "%dx", "%al", "%edi", "%ecx", "%eax", "%ebx"
   );
}



/* Atribui toda tela virtual com uma soh cor.
**********************************************************/
void CGr::mAtrVrCor ( BYTE yCor )
{
   memset ( (void*) m_payVr [ m_nVrAtiva ], yCor, TAMVR );
}



/* Marca um pixel na memoria virtual
**********************************************************/
void CGr::mAtrVrPixel ( INT nX, INT nY, BYTE yCor )
{
   memset ( (void*) m_payVr [ m_nVrAtiva ] + nX + ( nY * RESX ), yCor, 1 );
}



/* Transfere a memoria virtual para a tela, com sincronizacao
 de retraco. Ativa tela virtual inativa e desativa a ativa.
**********************************************************/
void CGr::mAtrVdVrSinc ()
{
   mVrSincRetraco ();
   
   #ifdef TRG_DOS_DJGPP

   // move dados de 4 em 4 bytes
   _movedatal ( _my_ds (), (INT) m_payVr [ m_nVrAtiva ], _dos_ds, 0xA0000, TAMVR4 );
   
   #endif
   
   #ifdef TRG_LINUX_SVGA
   
   // draw scanline per scanline
   for ( int ncL = 0; ncL < RESY; ncL ++ )
      vga_drawscansegment ( m_payVr [ m_nVrAtiva ] + ncL * RESX, 0, ncL, RESX );
   
   #endif

   m_nVrAtiva = ! m_nVrAtiva;
}



/* Desenha um poligono na memoria virtual de video usando
o incremento de x para calcular posicao de x.
Para garantir desenho de poligino convexo os vertices
adjacentes devem ocupar posicoes contiguas nos parametros
passado para essa funcao. Ou seja, visualmente, devem ser definidos
em ordem horaria ou anti-horaria. Exemplo:

                        (x4, y4)
                        *                -
                       / \                 loop
                      /   \                    |
            (x1, y1) *     \                   |
                      \     * (x3, y3)         | vai calculando posicoes de x
                       \   /                   | entre retas horizontais e
                        \ /                    V retas das bordas do poligono
                         *               -
                         (x2, y2)

Equacao das retas:
Y = aX + b, onde
            a eh declividade
            b eh deslocamento da reta no eixo y
Equacao usada nas interseccoes (x eh calculado definido quanto x varia a cada
linha e somando a um valor inicial de x):
X = ( deltaY * incrementoX ) + Xvertice, onde
            incrementoX eh 1 / a, quanto x varia por linha
            deltaY eh a distancia entre o Y de um vertice da reta e Y do loop
            Xvertice eh o X do vertice usado no calculo de deltaY

Os vertices sao numerados de 1 a 4 correspondendo a sua ordem nos parametros.
Sao chamados vertices superior ou inferior os vertices dos extremos
considerando y, no exemplo sao (x4, y4) e (x2, y2), respectivamente.

DESENHO DE LINHAS HORIZONTAIS:

Parte que desenha linha horizontal apenas onde nao foi desenhado (obsoleto):
// desenha linha horizontal, testa para nao desenhar por cima
__asm__ __volatile__ ("

      movl  %0,      %%edi   edi recebe a coordenada inicial (endereco inicial a ser zerado)
      movl  %1,      %%ecx   ecx recebe a coordenada final (ultimo endereco a ser zerado)
      movb  %2,      %%bh    bh recebe a cor (valor a ser escrito)

  Des:cmpb  $0,     (%%edi)  verifica se jah foi desenhado no byte (endereco apontado por edi tem valor == 0)
      jne   Inc              se jah foi desenhado nao marca ponto e passa para proxima coluna ((%%edi) != 0)
      movb  %%bh,   (%%edi)  marca ponto (escreve no endereco apontado por edi a cor - o conteudo de ebx)
  Inc:cmpl  %%edi,   %%ecx   verifica se terminou de tracar a linha (endereco atual igual ao final)
      je    Para             se terminou (edi == ecx) vai para fim
      incl  %%edi            vai para a proxima coluna (soma 1 byte no edi)
      jmp   Des              continua a desenhar
  Para:
      "
      :
      : "g" ( uEndLinha + nXIni ), "g" ( uEndLinha + nXFin ), "g" ( yCor )
      : "%bh", "%ecx", "%edi"
   );

Parte que desenha linha horizontal sem testar onde foi desenhado:
desenha linha horizontal byte a byte usando testes condicionais
__asm__ __volatile__ ("

    movl  %0,      %%edi
    movl  %1,      %%ecx
    movb  %2,      %%bh

Des:movb  %%bh,   (%%edi)
    cmpl  %%edi,   %%ecx
    je    Para
    incl  %%edi
    jmp   Des
Para:
   "
    :
    : "g" ( uEndLinha + nXIni ), "g" ( uEndLinha + nXFin ), "g" ( yCor )
    : "%bh", "%ecx", "%edi"
);
desenha linha horizontal byte a byte usando strings
__asm__ __volatile__ ("

      movl  %0,      %%edi    destino
      movl  %1,      %%ecx    numero de interacoes (comprimento da linha)
      movb  %2,      %%al     cor

      cld                     limpa flag de direcao
      rep                     inicia repeticao ateh ecx ser 0 (fim da linha)
         stosb                marca a memoria com o valor de al (cor)
      "
      :
      : "g" ( uEndLinha + nXIni ), "g" ( nXFin - nXIni ), "g" ( yCor )
      : "%al", "%ecx", "%edi"
   );
desenha linha horizontal de 4 em 4 bytes usando strings
__asm__ __volatile__ ("

      movl  %0,      %%edi    destino
      movl  %1,      %%ecx    numero de interacoes (comprimento da linha)
      movb  %2,      %%eax    cor (repete cor 4 vezes ao longo dos 32 bits)

                              Se comprimento nao eh multiplo de 4 deve
                              desenhar linha de 4 em 4 bytes e o resto.
                              Inicia desenhando resto.
      movb  %%cl,    %%bl     carrega bh com ultimos 8 bits para analisar
      andb  $3,      %%bl     seleciona apenas 2 ultimos bits, formam o resto
      cmpb  $3,      %%bl     verifica se resto eh 3, se sim desenha 3 pontos
      je    DesResto3         desvia para desenho
      cmpb  $2,      %%bl     verifica se resto eh 2, se sim desenha 2 pontos
      je    DesResto2         desvia para desenho
      cmpb  $1,      %%bl     verifica se resto eh 1, se sim desenha 1 ponto
      je    DesResto1         desvia para desenho
      jmp   DesTudo           se nao ocorreu desvio o resto eh 0, desvia
   DesResto3:                 se resto eh 3
      stosb
   DesResto2:                 se resto eh 2 ou 3
      stosb
   DesResto1:                 se resto eh 3 ou 2 ou 1
      stosb

   DesTudo:                   independente de resto desenha linha
      shrl  $2,     %%ecx     Remove parte das interacoes referentes ao resto
                              (divide por 4).
      cld                     limpa flag de direcao
      rep                     inicia repeticao ateh ecx ser 0 (fim da linha)
         stosl                marca a memoria com o valor de eax (cor)
      "
      :
      : "g" ( uEndLinha + nXIni ), "g" ( nXFin - nXIni + 1 ), "g" ( uCor )
      : "%bl", "%ecx", "%edi", "eax"
);

**********************************************************/
void CGr::mAtrVrPol2D ( INT nX1, INT nY1, INT nX2, INT nY2, INT nX3, INT nY3,
                        INT nX4, INT nY4, BYTE yCor )
{
   // recorta limites extremos de y
   // define valores dos extremos em y (maior e menor y)
   INT16
      nYSup = nY1,
      nYInf = nY1;
   // testa y2
   if ( nYSup > nY2 )
      nYSup = nY2;
   else
      if ( nYInf < nY2 )
         nYInf = nY2;
   // testa y3
   if ( nYSup > nY3 )
      nYSup = nY3;
   else
      if ( nYInf < nY3 )
         nYInf = nY3;
   // testa y4
   if ( nYSup > nY4 )
      nYSup = nY4;
   else
      if ( nYInf < nY4 )
         nYInf = nY4;
   // testa se extremos estao fora dos limites da tela, se sim define dentro
   if ( nYSup < MINY )
      nYSup = MINY;
   if ( nYInf > MAXY )
      nYInf = MAXY;

   // define se eh possivel calcular incrementos de x, se sim calcula
   INT
      nIncX12 = 0,
      nIncX14 = 0,
      nIncX23 = 0,
      nIncX34 = 0;
   BOOL
      bIncValido12 = TRUE,
      bIncValido14 = TRUE,
      bIncValido23 = TRUE,
      bIncValido34 = TRUE;
   // sao calculados valores apenas entre pontos que nao estao em linhas ==
   if ( nY1 != nY2 )
      nIncX12 = (INT) ( INCREMENTOX ( nX1, nY1, nX2, nY2 ) * 65536 );
   else
      bIncValido12 = FALSE;
   if ( nY1 != nY4 )
      nIncX14 = (INT) ( INCREMENTOX ( nX1, nY1, nX4, nY4 ) * 65536 );
   else
      bIncValido14 = FALSE;
   if ( nY2 != nY3 )
      nIncX23 = (INT) ( INCREMENTOX ( nX2, nY2, nX3, nY3 ) * 65536 );
   else
      bIncValido23 = FALSE;
   if ( nY3 != nY4 )
      nIncX34 = (INT) ( INCREMENTOX ( nX3, nY3, nX4, nY4 ) * 65536 );
   else
      bIncValido34 = FALSE;

   // calcula endereco da primeira linha de video que serah desenhada
   UINT
      uEndLinha = (UINT) m_payVr [ m_nVrAtiva ] + RESX * nYSup;

   // define cor de preenchimento
   UINT
      uCor = yCor | ( yCor << 8 ) | ( yCor << 16 ) | ( yCor << 24 );

   // varre poligono do extremo superior ateh o extremo inferior
   INT16
      nXIni,    // deve possuir apenas x inicial da linha horizontal
      nXFin,    // deve possuir apenas x final da linha horizontal
      nX;       // valor x da interceccao entre duas retas
   for ( INT16 ncY = nYSup; ncY <= nYInf; ncY ++ ) {

      // valores extremos iniciais de x (inicio e fim da linha horizontal)
      nXIni = 32767;
      nXFin = - 32768;

      // se estao sendo definidas posicoes de x entre vertices 1 e 4
      if ( bIncValido14 &&
           ( ncY >= nY1 && ncY <= nY4 ) ||
           ( ncY <= nY1 && ncY >= nY4 ) ) {

         nX = (INT16) ( ( ncY - nY4 ) * nIncX14 >> 16 ) + nX4;

         // verifica se eh para alterar x inicial
         if ( nX < nXIni )
            nXIni = nX;
         // verifica se eh para alterar x final
         if ( nX > nXFin )
            nXFin = nX;
      }

      // se estao sendo definidas posicoes de x entre vertices 1 e 2
      if ( bIncValido12 &&
           ( ncY >= nY1 && ncY <= nY2 ) ||
           ( ncY <= nY1 && ncY >= nY2 ) ) {

         nX = (INT16) ( ( ncY - nY2 ) * nIncX12 >> 16 ) + nX2;

         // verifica se eh para alterar x inicial
         if ( nX < nXIni )
            nXIni = nX;
         // verifica se eh para alterar x final
         if ( nX > nXFin )
            nXFin = nX;
      }

      // se estao sendo definidas posicoes de x entre vertices 2 e 3
      if ( bIncValido23 &&
           ( ncY >= nY2 && ncY <= nY3 ) ||
           ( ncY <= nY2 && ncY >= nY3 ) ){

         nX = (INT16) ( ( ncY - nY3 ) * nIncX23 >> 16 ) + nX3;

         // verifica se eh para alterar x inicial
         if ( nX < nXIni )
            nXIni = nX;
         // verifica se eh para alterar x final
         if ( nX > nXFin )
            nXFin = nX;
      }

      // se estao sendo definidas posicoes de x entre vertices 3 e 4
      if ( bIncValido34 &&
           ( ncY >= nY3 && ncY <= nY4 ) ||
           ( ncY <= nY3 && ncY >= nY4 ) ) {

         nX = (INT16) ( ( ncY - nY4 ) * nIncX34 >> 16 ) + nX4;

         // verifica se eh para alterar x inicial
         if ( nX < nXIni )
            nXIni = nX;
         // verifica se eh para alterar x final
         if ( nX > nXFin )
            nXFin = nX;
      }

      // Recorta valores extremos da linha horizontal (maior e menor x),
      // corrige valores dos extremos em x se estiverem fora da tela.
      // verifica inicio
      if ( nXIni < MINX )
         nXIni = MINX;
      else
         // inicio pode ser muito a direita: linha toda fora da tela
         if ( nXIni > MAXX ) {
            // calcula endereco da proxima linha de video
            uEndLinha += RESX;
            // nao desenha
            continue;
         }
      // verifica fim
      if ( nXFin > MAXX )
         nXFin = MAXX;
      else
         // fim pode ser muito a esquerda: linha toda fora da tela
         if ( nXFin < MINX ) {
            // calcula endereco da proxima linha de video
            uEndLinha += RESX;
            // nao desenha
            continue;
         }

      // desenha linha horizontal de 4 em 4 bytes
      __asm__ __volatile__ ("

            movl  %0,      %%edi
            movl  %1,      %%ecx
            movl  %2,      %%eax

            movb  %%cl,    %%bl
            andb  $3,      %%bl
            cmpb  $3,      %%bl
            je    DesResto3
            cmpb  $2,      %%bl
            je    DesResto2
            cmpb  $1,      %%bl
            je    DesResto1
            jmp   DesTudo
         DesResto3:
            stosb
         DesResto2:
            stosb
         DesResto1:
            stosb

         DesTudo:
            shrl  $2,      %%ecx
            cld
            rep
               stosl
            "
            :
            : "g" ( uEndLinha + nXIni ), "g" ( nXFin - nXIni + 1 ), "g" ( uCor )
            : "%bl", "%ecx", "%edi", "eax"
      );

      // calcula endereco da proxima linha de video (pode nao ser desenhada)
      uEndLinha += RESX;
   }
}



/* Copia uma area de memoria para a memoria virtual ativa
payImg - ponteiro para imagem; uIniX e uIniY - posicao inicial
de desenho da imagem; uDimX e uDimY - largura e altura da imagem.
Nao eh testado se imagem sai da tela.
**********************************************************/
void CGr::mAtrVrImagem ( BYTE* payImg, UINT16 uIniX, UINT16 uIniY,
                                       UINT16 uDimX, UINT16 uDimY )
{
   // copia linha por linha
   for ( UINT16 ncY = 0; ncY < uDimY; ncY ++ )
      memcpy ( (void*) ( m_payVr [ m_nVrAtiva ] + ( ncY + uIniY ) * RESX + uIniX ), (const void*) ( payImg + ncY * uDimX ), uDimX );
}
