/* Funcoes graficas basicas com acesso direto ao video.

Assembler AT&T + DJGPP:

- se gravar em variavel de memoria passada pelos parametros
  por "memory" no 3 :;
- __volatile__ permite ao compilador reorganizar o codigo */



#include <string.h>   // necessario para acessar sys/movedata.h
#include <stdio.h>
#include <stdlib.h>

#ifdef TRG_DOS_DJGPP
#include <dos.h>
#include <go32.h>
#include <conio.h>
#include <pc.h>
#endif

#ifdef TRG_LINUX_SVGA
#include <vga.h>
#endif

#include "defs.h"

#include "grvd.h"



/* construtor - inicia modo grafico, programa palette.
bInicGrVd == TRUE (default) inicia modo grafico, reprograma palette.
bInicGrVd altera comportamento do destrutor.
**********************************************************/
CGrVd::CGrVd ( BOOL bInicGrVd )
{
   // define como serah iniciada a classe
   m_bInicGrVd = bInicGrVd;

   // verifica se deve entrar em modo grafico e reprogramar palette
   if ( m_bInicGrVd == TRUE ) {

      // entra em modo grafico
      mAtrVdModo ( MCGA );

      // programa palette
      // palette eh dividida em 8 blocos de tons de cores diferentes
      for ( INT ncBase = 0; ncBase < 8; ncBase ++ ) {
         BYTE
            yTom = ILUMAMB;     // o numero eh o tom mais escuro usado
         // programa cada bloco com variacoes de tom da mesma cor
         for ( INT ncC = ncBase * NUMCOR; ncC < ( ncBase * NUMCOR + NUMCOR ); ncC ++ ) {

            // seleciona cor base
            switch ( ncBase ) {
               case 0:    // 0 - indice da palette
                  mAtrPalCor ( ncC, yTom, yTom, yTom );
                  break;
               case 1:    // 32
                  mAtrPalCor ( ncC, yTom, yTom, 5 );
                  break;
               case 2:    // 64
                  mAtrPalCor ( ncC, 5, 5, yTom );
                  break;
               case 3:    // 96
                  mAtrPalCor ( ncC, yTom, 10, 10 );
                  break;
               case 4:    // 128
                  mAtrPalCor ( ncC, 2, yTom - 3, 2 );
                  break;
               case 5:    // 160
                  mAtrPalCor ( ncC, yTom - 2, yTom + 5, 25 );
                  break;
               case 6:    // 192
                  mAtrPalCor ( ncC, yTom - 20, yTom - 20, yTom - 20 );
                  break;
               case 7:    // 224
                  mAtrPalCor ( ncC, yTom, yTom - 10, yTom + 10 );
                  break;
            }
            yTom ++;
         }
      }
      mAtrPalCor ( 0, 0, 0, 0 );
      mAtrPalCor ( 2, 53, 39, 26 );
      mAtrPalCor ( 7, 45, 30, 20 );
   }
}



/* destrutor - Volta ao modo texto, desaloca memoria
**********************************************************/
CGrVd::~CGrVd ()
{
   // verifica se entrou em modo grafico no inicio da classe
   if ( m_bInicGrVd == TRUE )
      // entra em modo texto
      mAtrVdModo ( TEXTO );
}



/* Altera o modo de exibicao do video.
**********************************************************/
void CGrVd::mAtrVdModo ( UINT16 uModo )
{
   #ifdef TRG_DOS_DJGPP

   REGS
      stRegsE, stRegsS;      // registradores

   stRegsE.x.ax = uModo;
   int86 ( 0x10, &stRegsE, &stRegsS );

   // se entra em modo MCGA altera proporcao entre x e y para 1x1
   if ( uModo == MCGA )
      outportb ( 0x3C2, 0xE3 );

   #endif

   #ifdef TRG_LINUX_SVGA

   if ( uModo == MCGA ) {
      vga_init ();
      vga_setmode ( G320x200x256 );
   }
   else
      vga_setmode ( TEXT );
   
   #endif
}



/* Atribui uma quantidade de RGB para uma cor (entrada) da palette
**********************************************************/
void CGrVd::mAtrPalCor ( BYTE yCor, BYTE yR, BYTE yG, BYTE yB )
{
   #ifdef TRG_DOS_DJGPP

   // solicita alteracao de palette
   outportb ( 0x03C8, yCor );
   // altera
   outportb ( 0x03C9, yR );
   outportb ( 0x03C9, yG );
   outportb ( 0x03C9, yB );

   #endif

   #ifdef TRG_LINUX_SVGA

   vga_setpalette ( yCor, yR, yG, yB );

   #endif
}



/* Retorna os componentes RGB de uma cor (entrada) da palette
**********************************************************/
void CGrVd::mRetPalCor ( BYTE yCor, BYTE& yR, BYTE& yG, BYTE& yB )
{
   #ifdef TRG_DOS_DJGPP

   // solicita leitura de palette
   outportb ( 0x03C7, yCor );
   // le
   yR = inportb ( 0x03C9 );
   yG = inportb ( 0x03C9 );
   yB = inportb ( 0x03C9 );

   #endif

   #ifdef TRG_LINUX_SVGA

   INT
      nR = yR,
      nG = yG,
      nB = yB;
      
   vga_getpalette ( yCor, &nR, &nG, &nB );
   
   yR = nR;
   yG = nG;
   yB = nB;
   
   #endif
}



/* Guarda em tabelas internas os componentes RGB de todas as
 cores da palette. Usado para retaurar palette apohs alteracoes
 com CGrVd::mRestauraPal
**********************************************************/
void CGrVd::mSalvaPal ()
{
   // varre toda palette lendo componentes de cada cor
   for ( UINT16 ucCor = 0; ucCor < 256; ucCor ++ )
      mRetPalCor ( ucCor, m_ayPalR [ ucCor ], m_ayPalG [ ucCor ], m_ayPalB [ ucCor ] );
}



/* Altera a palette com as cores das tabelas internas. Usado em
conjunto com CGrVd::mSalvaPal
**********************************************************/
void CGrVd::mRestauraPal ()
{
   // varre toda palette alterando componentes de cada cor
   for ( UINT16 ucCor = 0; ucCor < 256; ucCor ++ )
      mAtrPalCor ( ucCor, m_ayPalR [ ucCor ], m_ayPalG [ ucCor ], m_ayPalB [ ucCor ] );
}



/* Programa a palette conforme componentes RGB passadas por uma tabela
 cada 3 bytes da tabelainformam uma cor.
**********************************************************/
void CGrVd::mAtrPal ( BYTE* ayPalRGB )
{
   // varre toda palette alterando componentes de cada cor
   for ( UINT16 ucCor = 0; ucCor < 768; ucCor += 3 )
      mAtrPalCor ( ucCor / 3, * ( ayPalRGB + ucCor ), * ( ayPalRGB + ucCor + 1 ), * ( ayPalRGB + ucCor + 2 ) );
}



/* Espera retraco de video para desenhar (=="WaitRetrace"),
**********************************************************/
void CGrVd::mVdSincRetraco ()
{
   #ifdef TRG_DOS_DJGPP

   __asm__ __volatile__ ("

          movw  $0x03DA, %%dx
      L1: inb   %%dx,    %%al
          andb  $0x8,    %%al
          jnz   L1

      L2: inb   %%dx,    %%al
          andb  $0x8,    %%al
          jz    L2
      "
      :
      :
      : "%dx", "%al"
   );

   #endif

   #ifdef TRG_LINUX_SVGA

   vga_waitretrace ();
      
   #endif
}



/* Atribui toda a tela com uma soh cor. Sincroniza com retraco.
**********************************************************/
void CGrVd::mAtrVdCorSinc ( BYTE yCor )
{
   BYTE
      ayCores [ 4 ];
   ayCores [ 0 ] = ayCores [ 1 ] = ayCores [ 2 ] = ayCores [ 3 ] = yCor;

   mVdSincRetraco ();

   #ifdef TRG_DOS_DJGPP

   for ( INT ncTrans = 0; ncTrans < TAMVR; ncTrans += 4 )
      _dosmemputl ( (void*) ayCores, 1, 0xA0000 + ncTrans );

   #endif

   #ifdef TRG_LINUX_SVGA

   vga_setcolor ( yCor );
   for ( INT ncL = 0; ncL < RESY; ncL ++ )
      vga_drawline ( 0, ncL, MAXX, ncL );
      
      
   #endif
}



/* Marca um pixel na tela
**********************************************************/
void CGrVd::mAtrVdPixel ( INT nX, INT nY, BYTE yCor )
{

   #ifdef TRG_DOS_DJGPP

   dosmemput ( (void*) &yCor, 1, 0xA0000 + nX + ( nY * RESX ) );

   #endif

   #ifdef TRG_LINUX_SVGA

   vga_setcolor ( yCor );
   vga_drawpixel (  nX, nY );
      
   #endif
}
